import json
import time
import random
from dcicutils import ff_utils
from dcicutils.s3_utils import s3Utils
from dcicutils.env_utils_legacy import FF_PRODUCTION_IDENTIFIER, FF_STAGING_IDENTIFIER
from datetime import datetime, timezone, timedelta
from operator import itemgetter
from tibanna_4dn.core import API
from . import wfrset_utils

lambda_limit = wfrset_utils.lambda_limit
random_wait = wfrset_utils.random_wait
load_wait = wfrset_utils.load_wait


# wfr_name, accepted versions, expected run time # wfr_name, accepted versions,
workflow_details = {
    "md5": {
        "run_time": 12,
        "accepted_versions": ["0.0.4", "0.2.6"]
    },
    # old workflow run naming, updated on workflows for old ones
    "fastqc-0-11-4-1": {
        "run_time": 50,
        "accepted_versions": ["0.2.0"]
    },
    "fastqc": {
        "run_time": 50,
        "accepted_versions": ["v1", "v2"]
    },
    "bwa-mem": {
        "run_time": 50,
        "accepted_versions": ["0.2.6"]
    },
    "pairsqc-single": {
        "run_time": 100,
        "accepted_versions": ["0.2.5", "0.2.6"]
    },
    "hi-c-processing-bam": {
        "run_time": 200,
        "accepted_versions": ["0.2.6"]
    },
    "hi-c-processing-pairs": {
        "run_time": 200,
        "accepted_versions": ["0.2.6", "0.2.7"]
    },
    "hi-c-processing-pairs-nore": {
        "run_time": 200,
        "accepted_versions": ["0.2.6"]
    },
    "hi-c-processing-pairs-nonorm": {
        "run_time": 200,
        "accepted_versions": ["0.2.6"]
    },
    "hi-c-processing-pairs-nore-nonorm": {
        "run_time": 200,
        "accepted_versions": ["0.2.6"]
    },
    "repliseq-parta": {
        "run_time": 200,
        "accepted_versions": ["v13.1", "v14", "v16"]
    },
    "bedGraphToBigWig": {
        "run_time": 24,
        "accepted_versions": ["v4", "v5", "v6"]
    },
    "bedtomultivec": {
        "run_time": 24,
        "accepted_versions": ["v4"]
    },
    "bedtobeddb": {
        "run_time": 24,
        "accepted_versions": ["v2", "v3"]
    },
    "encode-chipseq-aln-chip": {
        "run_time": 200,
        "accepted_versions": ["1.1.1"]
    },
    "encode-chipseq-aln-ctl": {
        "run_time": 200,
        "accepted_versions": ["1.1.1"]
    },
    "encode-chipseq-postaln": {
        "run_time": 200,
        "accepted_versions": ["1.1.1"]
    },
    "encode-atacseq-aln": {
        "run_time": 200,
        "accepted_versions": ["1.1.1"]
    },
    "encode-atacseq-postaln": {
        "run_time": 200,
        "accepted_versions": ["1.1.1"]
    },
    "mergebed": {
        "run_time": 200,
        "accepted_versions": ["v1"]
    },
    'imargi-processing-fastq': {
        "run_time": 50,
        "accepted_versions": ["1.1.1_dcic_4"]
    },
    'imargi-processing-bam': {
        "run_time": 50,
        "accepted_versions": ["1.1.1_dcic_4"]
    },
    'imargi-processing-pairs': {
        "run_time": 200,
        "accepted_versions": ["1.1.1_dcic_4"]
    },
    'encode-rnaseq-stranded': {
        "run_time": 200,
        "accepted_versions": ["1.1"]
    },
    'encode-rnaseq-unstranded': {
        "run_time": 200,
        "accepted_versions": ["1.1"]
    },
    'rna-strandedness': {
        "run_time": 200,
        "accepted_versions": ["v2"]
    },
    'bamqc': {
        "run_time": 200,
        "accepted_versions": ["v2", "v3"]
    },
    'fastq-first-line': {
        "run_time": 200,
        "accepted_versions": ["v2"]
    },
    're_checker_workflow': {
        "run_time": 200,
        "accepted_versions": ['v1.1', 'v1.2']
    },
    'mad_qc_workflow': {
        "run_time": 200,
        "accepted_versions": ['1.1_dcic_2']
    },
    'merge-fastq': {
        "run_time": 200,
        "accepted_versions": ['v1']
    },
    'insulation-scores-and-boundaries-caller': {
            "run_time": 200,
            "accepted_versions": ['v1']
    },
    'compartments-caller': {
                "run_time": 200,
                "accepted_versions": ['v1.2']
    },
    'mcoolQC': {
                "run_time": 200,
                "accepted_versions": ['v1']
    }
}


# accepted versions for completed pipelines
accepted_versions = {
    # OFFICIAL
    'in situ Hi-C':  ["HiC_Pipeline_0.2.6", "HiC_Pipeline_0.2.7"],
    # OFFICIAL
    'Dilution Hi-C': ["HiC_Pipeline_0.2.6", "HiC_Pipeline_0.2.7"],
    # OFFICIAL
    'TCC':           ["HiC_Pipeline_0.2.6", "HiC_Pipeline_0.2.7"],
    # OFFICIAL  # NO-RE
    'DNase Hi-C':    ["HiC_Pipeline_0.2.6", "HiC_Pipeline_0.2.7"],
    # OFFICIAL  # NO-NORM
    'Capture Hi-C':  ["HiC_Pipeline_0.2.6", "HiC_Pipeline_0.2.7"],
    # OFFICIAL  # NO-RE
    'Micro-C':       ["HiC_Pipeline_0.2.6", "HiC_Pipeline_0.2.7"],
    # Preliminary - Released to network  # NO-RE NO-NORM
    'ChIA-PET':      ["HiC_Pipeline_0.2.6", "HiC_Pipeline_0.2.7"],
    # Preliminary - Released to network  # NO-RE NO-NORM
    'in situ ChIA-PET': ["HiC_Pipeline_0.2.7"],
    # Preliminary - Released to network  # NO-RE NO-NORM
    'TrAC-loop':     ["HiC_Pipeline_0.2.6", "HiC_Pipeline_0.2.7"],
    # Preliminary - Released to network  # NO-NORM
    'PLAC-seq':      ["HiC_Pipeline_0.2.6", "HiC_Pipeline_0.2.7"],
    # Preliminary - Released to network  # NO-NORM
    'HiChIP': ["HiC_Pipeline_0.2.7"],
    # bwa mem # handled manually for now
    'MARGI':         ['MARGI_Pipeline_1.1.1_dcic_4'],
    # Preliminary -  Don't release - (Released to network is pending approval from Belmont lab)
    'TSA-seq':       ['RepliSeq_Pipeline_v13.1_step1',
                      'RepliSeq_Pipeline_v14_step1',
                      'RepliSeq_Pipeline_v16_step1'],
    # OFFICIAL - 1 STEP
    '2-stage Repli-seq': ['RepliSeq_Pipeline_v13.1_step1',
                          'RepliSeq_Pipeline_v14_step1',
                          'RepliSeq_Pipeline_v16_step1'],
    # OFFICIAL - 1 STEP
    'Multi-stage Repli-seq': ['RepliSeq_Pipeline_v13.1_step1',
                              'RepliSeq_Pipeline_v14_step1',
                              'RepliSeq_Pipeline_v16_step1'],
    # Preliminary - Released to network
    'NAD-seq':       ['RepliSeq_Pipeline_v13.1_step1', 'RepliSeq_Pipeline_v14_step1', 'RepliSeq_Pipeline_v16_step1'],
    # OFFICIAL
    'ATAC-seq':      ['ENCODE_ATAC_Pipeline_1.1.1'],
    # OFFICIAL
    'ChIP-seq':      ['ENCODE_ChIP_Pipeline_1.1.1'],
    # OFFICIAL
    'RNA-seq': ['ENCODE_RNAseq_Pipeline_1.1'],
    'single cell Repli-seq': [''],
    'cryomilling TCC': [''],
    'single cell Hi-C': [''],
    'sci-Hi-C': [''],
    'MC-3C': [''],
    'MC-Hi-C': [''],
    'DamID-seq': [''],
    'DNA SPRITE': [''],
    'RNA-DNA SPRITE': [''],
    'GAM': [''],
    'CUT&RUN': [''],
    'TRIP': ['']
    }

# Accepted versions for feature calling pipelines
feature_calling_accepted_versions = {
    'insulation_scores_and_boundaries': ["insulation_scores_and_boundaries_v1"],
    'compartments': ["compartments_v1.2"]
}
# Reference Files
bwa_index = {"human": "4DNFIZQZ39L9",
             "mouse": "4DNFI823LSI8",
             "fruit-fly": '4DNFIO5MGY32',
             "chicken": "4DNFIVGRYVQF",
             "zebrafish": "4DNFIUH46PG1"}

chr_size = {"human": "4DNFI823LSII",
            "mouse": "4DNFI3UBJ3HZ",
            "fruit-fly": '4DNFIBEEN92C',
            "chicken": "4DNFIQFZW4DX",
            "zebrafish": "4DNFI5W8CN1M"}

# star index for rna Seq
rna_star_index = {"human": "4DNFI3FCGSW2",
                  "mouse": "4DNFINJIU765"}

# star index for rna Seq
rna_rsem_index = {"human": "4DNFIB4HV398",
                  "mouse": "4DNFI2GFI8KN"}

# chromosome sizes for rna Seq
rna_chr_size = {"human": "4DNFIZJB62D1",
                "mouse": "4DNFIBP173GC"}

# id to gene type for rna Seq
rna_t2g = {"human": "4DNFIHBJI984",
           "mouse": "4DNFINBJ25DT"}

re_nz = {"human": {'MboI': '/files-reference/4DNFI823L812/',
                   'DpnII': '/files-reference/4DNFIBNAPW3O/',
                   'HindIII': '/files-reference/4DNFI823MBKE/',
                   'NcoI': '/files-reference/4DNFI3HVU2OD/',
                   'MspI': '/files-reference/4DNFI2JHR3OI/',
                   'NcoI_MspI_BspHI': '/files-reference/4DNFI6HA6EH9/',
                   'AluI': '/files-reference/4DNFIN4DB5O8/',
                   'DdeI': '/files-reference/4DNFI4YGL4RE/',
                   'DdeI and DpnII': '/files-reference/4DNFIS1FCRRK/',
                   'MseI': '/files-reference/4DNFIMD6BNQ8/',
                   "Arima - A1, A2": '/files-reference/4DNFIU1RS39P/'
                   },
         "mouse": {'MboI': '/files-reference/4DNFIONK4G14/',
                   'DpnII': '/files-reference/4DNFI3HVC1SE/',
                   "HindIII": '/files-reference/4DNFI6V32T9J/',
                   "Arima - A1, A2": '/files-reference/4DNFIE78H3K7/'
                   },
         "fruit-fly": {'MboI': '/files-reference/4DNFIS1ZVUWO/'
                       },
         "chicken": {"HindIII": '/files-reference/4DNFITPCJFWJ/'
                     },
         "zebrafish": {'MboI': '/files-reference/4DNFI6OUDFWL/'
                       }
         }


# re bed files for MARGI pipeline
re_fragment = {"human": {'AluI': '/files-reference/4DNFIL1I5TSP/'}}

# reference beta-actin 21kmer for rna_strandness pipeline
re_kmer = {"human": '/files-reference/4DNFIDMVPFSO/',
           "mouse": '/files-reference/4DNFIAQ4BI8Y'}

# max_distance for species (used for pairsqc)
max_size = {"human": None,
            "mouse": 8.2,
            "fruit-fly": 7.5,
            "chicken": 8.2,
            "zebrafish": 7.9}

# Restriction enzyme recognition site length`
re_nz_sizes = {"HindIII": "6",
               "DpnII": "4",
               "MboI": "4",
               "NcoI": "6",
               "MspI": "4",
               "BspHI": "6",
               "DdeI and DpnII": "4",
               "DdeI": "4",
               "NcoI_MspI_BspHI": "4",  # this is an NZ mix, no of cut sites should be similar to 4 cutter mspI
               "MseI": "4",
               "Arima - A1, A2": "4"  # this is an NZ mix, no of cut sites should be similar to 4 cutter mspI
               }

mapper = {'human': 'GRCh38',
          'mouse': 'GRCm38',
          'fruit-fly': 'dm6',
          'chicken': 'galGal5',
          'zebrafish': 'GRCz11'}

# color map states bed file
states_file_type = {
    'SPIN_states_v1':
        {
            'color_mapper': '/files-reference/4DNFI27WSLAG/',
            'num_states': 9
            }
        }


# GC% content reference files (compartments pipeline)
gc_content_ref = {"human": "/files-reference/4DNFI7MCA4R6/",
                  "mouse": "/files-reference/4DNFIOFXJOUA",
                  'fruit fly': "/files-reference/4DNFID6KQ941",
                  "chicken": "/files-reference/4DNFI19V162N",
                  "zebrafish": "/files-reference/4DNFIHEHIZ3P"}


def check_indexing(check, connection):
    """Checks the indexing queue, if there are items in the queue,
    Modifies the check, and returns it along with a flag that is set to True,
    if no items, returns original checks, and flag False"""
    # wait for random time
    wait = round(random.uniform(0.1, random_wait), 1)
    time.sleep(wait)
    # # TEMPORARILY DISABLE ALL PIPELINE RUNS
    # check.status = 'PASS'  # maybe use warn?
    # check.brief_output = ['Check Temporarily Disabled']
    # check.summary = 'Check Temporarily Disabled'
    # check.full_output = {}
    # return check, True
    # check indexing queue
    env = connection.ff_env
    if env in [FF_PRODUCTION_IDENTIFIER, FF_STAGING_IDENTIFIER]:
        health = ff_utils.get_health_page(ff_env=env)
        env = health['beanstalk_env']  # this is ENV_NAME and needs to match to get the correct queue
    indexing_queue = ff_utils.stuff_in_queues(env, check_secondary=True)
    if indexing_queue:
        check.status = 'PASS'  # maybe use warn?
        check.brief_output = ['Waiting for indexing queue to clear']
        check.summary = 'Waiting for indexing queue to clear'
        check.full_output = {}
        return check, True
    else:
        return check, False


# check for a specific tag in a states file
def isthere_states_tag(a_file):
    if a_file.get('tags'):
        for tag in a_file['tags']:
            if tag not in states_file_type:
                return (False, 'unregistered_tag')
            else:
                return (True, '')
    else:
        return (False, 'missing_tag')


def extract_nz_chr(acc, auth):
    """Get RE nz recognition site length and chrsize file accession
    used for pairsqc."""
    exp_resp = ff_utils.get_metadata(acc, key=auth)
    exp_type = exp_resp['experiment_type']['display_title']
    # get enzyme
    nz_num = ""
    nz = exp_resp.get('digestion_enzyme')
    if nz:
        nz_num = re_nz_sizes.get(nz['display_title'])
    if nz_num:
        pass
    # Use 6 for Chiapet and all without nz (Soo & Burak)
    elif exp_type in ['in situ ChIA-PET', 'ChIA-PET', 'Micro-C', 'DNase Hi-C', 'TrAC-loop']:
        nz_num = '6'
    else:
        return (None, None, 'No enzyme or accepted exp type')
    # get organism
    biosample = exp_resp['biosample']
    organisms = list(set([bs['organism']['name'] for bs in biosample['biosource']]))
    chrsize = ''
    if len(organisms) == 1:
        chrsize = chr_size.get(organisms[0])
    else:
        # multiple organism biosample
        return (None, None, 'Biosample contains multiple organism')
    # if organism is not available return empty
    if not chrsize:
        msg = organisms[0] + ' does not have chrsize file'
        return (None, None, msg)
    # organism should be in max size dict
    assert organisms[0] in max_size
    max_distance = max_size.get(organisms[0])
    # return result if both exist
    return nz_num, chrsize, max_distance


def check_qcs_on_files(file_meta, all_qcs):
    """Go over qc related fields, and check for overall quality score."""
    def check_qc(file_accession, resp, failed_qcs_list):
        """format errors and return a errors list."""
        quality_score = resp.get('overall_quality_status', '')
        if quality_score.upper() != 'PASS':
            failed_qcs_list.append([file_accession, resp['display_title'], resp['uuid']])
        return failed_qcs_list

    failed_qcs = []
    if not file_meta.get('quality_metric'):
        return
    qc_result = [i for i in all_qcs if i['@id'] == file_meta['quality_metric']['@id']][0]
    if qc_result['display_title'].startswith('QualityMetricQclist'):
        if not qc_result.get('qc_list'):
            return
        for qc in qc_result['qc_list']:
            qc_resp = [i for i in all_qcs if i['@id'] == qc['@id']][0]
            failed_qcs = check_qc(file_meta['accession'], qc_resp, failed_qcs)
    else:
        failed_qcs = check_qc(file_meta['accession'], qc_result, failed_qcs)
    return failed_qcs


def stepper(library, keep,
            step_tag, sample_tag, new_step_input_file,
            input_file_dict,  new_step_name, new_step_output_arg,
            additional_input={}, organism='human', no_output=False):
    """This functions packs the core of wfr check, for a given workflow and set of
    input files, it will return the status of process on these files.
    It will also check for failed qcs on input files.
    new_step_output_arg= can be str or list, will return str or list of @id for output files with given argument(s)"""
    step_output = ''
    # unpack library
    all_files = library['files']
    all_wfrs = library['wfrs']
    all_qcs = library['qcs']
    # unpack keep
    running = keep['running']
    problematic_run = keep['problematic_run']
    missing_run = keep['missing_run']

    # Lets get the repoinse from one of the input files that will be used in this step
    # if it is a list take the first item, if not use it as is
    # new_step_input_file must be the @id
    # also check for qc status
    qc_errors = []
    if isinstance(new_step_input_file, list) or isinstance(new_step_input_file, tuple):
        file_accs = []
        for an_input in new_step_input_file:
            # for chip seq we need another level of unwrapping
            if isinstance(an_input, list) or isinstance(an_input, tuple):
                for a_nested_input in an_input:
                    file_accs.append(a_nested_input.split('/')[2])
                    input_resp = [i for i in all_files if i['@id'] == a_nested_input][0]
                    errors = check_qcs_on_files(input_resp, all_qcs)
                    if errors:
                        qc_errors.extend(errors)
            else:
                file_accs.append(an_input.split('/')[2])
                input_resp = [i for i in all_files if i['@id'] == an_input][0]
                errors = check_qcs_on_files(input_resp, all_qcs)
                if errors:
                    qc_errors.extend(errors)
        name_tag = '_'.join(file_accs)
    else:
        input_resp = [i for i in all_files if i['@id'] == new_step_input_file][0]
        errors = check_qcs_on_files(input_resp, all_qcs)
        if errors:
            qc_errors.extend(errors)
        name_tag = new_step_input_file.split('/')[2]
    # if there are qc errors, return with qc qc_errors
    if qc_errors:
        problematic_run.append([step_tag + ' input file qc error', qc_errors])
        step_status = "no complete run, qc error"
    # if no qc problem, go on with the run check
    else:
        if no_output:
            step_result = get_wfr_out(input_resp, new_step_name, all_wfrs=all_wfrs, md_qc=True)
        else:
            step_result = get_wfr_out(input_resp, new_step_name, all_wfrs=all_wfrs)
        step_status = step_result['status']
        # if successful
        input_file_accession = input_resp['accession']
        if step_status == 'complete':
            if new_step_output_arg:
                if isinstance(new_step_output_arg, list):
                    output_list = []
                    for an_output_arg in new_step_output_arg:
                        output_list.append(step_result[an_output_arg])
                    step_output = output_list
                else:
                    step_output = step_result[new_step_output_arg]
            pass
        # if still running
        elif step_status == 'running':
            running.append([step_tag, sample_tag, input_file_accession])
        # if run is not successful
        elif step_status.startswith("no complete run, too many"):
            problematic_run.append([step_tag, sample_tag, input_file_accession])
        else:
            # add step 4
            missing_run.append([step_tag, [new_step_name, organism, additional_input], input_file_dict, name_tag])

    keep['running'] = running
    keep['problematic_run'] = problematic_run
    keep['missing_run'] = missing_run
    return keep, step_status, step_output


def get_wfr_out(emb_file, wfr_name, key=None, all_wfrs=None, versions=None,
                md_qc=False, run=None, error_threshold=2):
    """For a given file, fetches the status of last wfr (of wfr_name type)
    If there is a successful run, it will return the output files as a dictionary of
    argument_name:file_id, else, will return the status. Some runs, like qc and md5,
    does not have any file_format output, so they will simply return 'complete'
    args:
     emb_file: embedded frame file info
     wfr_name: base name without version
     key: authorization
     all_wfrs : all releated wfrs in embedded frame
     versions: acceptable versions for wfr
     md_qc: if no output file is excepted, set to True
     run: if run is still running beyond this hour limit, assume problem
     error_threshold = if there are this many failed runs, don't proceed
    """
    # tag as problematic if problematic runs are this many
    # if there are n failed runs, don't proceed

    error_at_failed_runs = error_threshold
    # you should provide key or all_wfrs
    assert key or all_wfrs
    if wfr_name not in workflow_details:
        assert wfr_name in workflow_details
    # get default accepted versions if not provided
    if not versions:
        versions = workflow_details[wfr_name]['accepted_versions']
    # get default run out time
    if not run:
        run = workflow_details[wfr_name]['run_time']

    workflows = emb_file.get('workflow_run_inputs')
    wfr = {}
    run_status = 'did not run'
    if not workflows:
        return {'status': "no workflow on file"}
    my_workflows = [i for i in workflows if i['display_title'].startswith(wfr_name)]
    if not my_workflows:
        return {'status': "no workflow on file"}

    for a_wfr in my_workflows:
        wfr_type, time_info = a_wfr['display_title'].split(' run ')
        wfr_type_base, wfr_version = wfr_type.strip().split(' ')
        # user submitted ones use run on insteand of run
        time_info = time_info.strip('on').strip()
        try:
            wfr_time = datetime.strptime(time_info, '%Y-%m-%d %H:%M:%S.%f')
        except ValueError:
            wfr_time = datetime.strptime(time_info, '%Y-%m-%d %H:%M:%S')
        a_wfr['run_hours'] = (datetime.utcnow() - wfr_time).total_seconds() / 3600
        a_wfr['run_type'] = wfr_type_base.strip()
        a_wfr['run_version'] = wfr_version.strip()
    my_workflows = [i for i in my_workflows if i['run_version'] in versions]
    if not my_workflows:
        return {'status': "no workflow in file with accepted version"}
    my_workflows = sorted(my_workflows, key=lambda k: k['run_hours'])
    same_type_wfrs = [i for i in my_workflows if i['run_type'] == wfr_name]

    if not same_type_wfrs:
        return {'status': "no workflow on file"}
    last_wfr = same_type_wfrs[0]

    # get metadata for the last wfr
    if all_wfrs:
        wfr = [i for i in all_wfrs if i['uuid'] == last_wfr['uuid']][0]
    else:
        wfr = ff_utils.get_metadata(last_wfr['uuid'], key)
    run_duration = last_wfr['run_hours']
    run_status = wfr['run_status']

    if run_status == 'complete':
        outputs = wfr.get('output_files')
        # some runs, like qc, don't have a real file output
        if md_qc:
            return {'status': 'complete'}
        # if expected output files, return a dictionary of argname:file_id
        else:
            out_files = {}
            for output in outputs:
                if output.get('format'):
                    # get the arg name
                    arg_name = output['workflow_argument_name']
                    try:
                        out_files[arg_name] = output['value']['@id']
                    except KeyError:
                        out_files[arg_name] = None
            if out_files:
                out_files['status'] = 'complete'
                return out_files
            else:
                return {'status': "no file found"}
    # if status is error
    elif run_status == 'error':
        # are there too many failed runs
        if len(same_type_wfrs) >= error_at_failed_runs:
            return {'status': "no complete run, too many errors"}

        return {'status': "no complete run, errrored"}
    # if other statuses, started running
    elif run_duration < run:
        return {'status': "running"}
    # this should be the timeout case
    else:
        if len(same_type_wfrs) >= error_at_failed_runs:
            return {'status': "no complete run, too many time-outs"}
        else:
            return {'status': "no completed run, time-out"}


def get_attribution(file_json):
    """give file response in embedded frame and extract attribution info"""
    dciclab = '/labs/4dn-dcic-lab/'
    attributions = {
        'lab': dciclab,
        'award': '/awards/2U01CA200059-06/'
    }
    assert file_json.get('lab').get('@id')  # assume this is not really necessary
    file_lab = file_json['lab']['@id']
    cont_labs = []
    if file_json.get('contributing_labs'):
        cont_labs = [i['@id'] for i in file_json['contributing_labs']]
    if file_lab != dciclab and file_lab not in cont_labs:
        cont_labs.append(file_lab)
    if cont_labs:
        attributions['contributing_labs'] = cont_labs
    return attributions


def extract_file_info(obj_id, arg_name, additional_parameters, auth, env, rename=[]):
    """Takes file id, and creates info dict for tibanna"""
    my_s3_util = s3Utils(env=env)
    raw_bucket = my_s3_util.raw_file_bucket
    out_bucket = my_s3_util.outfile_bucket
    """Creates the formatted dictionary for files.
    """
    # start a dictionary
    template = {"workflow_argument_name": arg_name}
    if rename:
        change_from = rename[0]
        change_to = rename[1]
    # if it is list of items, change the structure
    # with chip seq, files might be wrapped in 3 levels of nesting
    # ToDo: an iterative unwrapper would simplify the code here
    if isinstance(obj_id, list):
        object_key = []
        uuid = []
        buckets = []
        print('1', obj_id)
        for obj in obj_id:
            print('2', obj)
            if isinstance(obj, (list, tuple)):
                nested_object_key = []
                nested_uuid = []
                for nested_obj in obj:
                    print('3', nested_obj)
                    if isinstance(nested_obj, (list, tuple)):
                        nested_nested_object_key = []
                        nested_nested_uuid = []
                        for nested_nested_obj in nested_obj:
                            print('4', nested_nested_obj)
                            metadata = ff_utils.get_metadata(nested_nested_obj, key=auth)
                            nested_nested_object_key.append(metadata['display_title'])
                            nested_nested_uuid.append(metadata['uuid'])
                            # get the bucket
                            if 'FileProcessed' in metadata['@type']:
                                my_bucket = out_bucket
                            else:  # covers cases of FileFastq, FileReference, FileMicroscopy
                                my_bucket = raw_bucket
                            buckets.append(my_bucket)
                        nested_object_key.append(nested_nested_object_key)
                        nested_uuid.append(nested_nested_uuid)
                    else:
                        metadata = ff_utils.get_metadata(nested_obj, key=auth)
                        nested_object_key.append(metadata['display_title'])
                        nested_uuid.append(metadata['uuid'])
                        # get the bucket
                        if 'FileProcessed' in metadata['@type']:
                            my_bucket = out_bucket
                        else:  # covers cases of FileFastq, FileReference, FileMicroscopy
                            my_bucket = raw_bucket
                        buckets.append(my_bucket)
                object_key.append(nested_object_key)
                uuid.append(nested_uuid)
            else:
                metadata = ff_utils.get_metadata(obj, key=auth)
                object_key.append(metadata['display_title'])
                uuid.append(metadata['uuid'])
                # get the bucket
                if 'FileProcessed' in metadata['@type']:
                    my_bucket = out_bucket
                else:  # covers cases of FileFastq, FileReference, FileMicroscopy
                    my_bucket = raw_bucket
                buckets.append(my_bucket)
        # check bucket consistency
        assert len(list(set(buckets))) == 1
        template['uuid'] = uuid
        if rename:
            template['rename'] = [i.replace(change_from, change_to) for i in object_key]
        if additional_parameters:
            template.update(additional_parameters)

    # if obj_id is a string
    else:
        metadata = ff_utils.get_metadata(obj_id, key=auth)
        template['uuid'] = metadata['uuid']
        # get the bucket
        if 'FileProcessed' in metadata['@type']:
            my_bucket = out_bucket
        else:  # covers cases of FileFastq, FileReference, FileMicroscopy
            my_bucket = raw_bucket
        if rename:
            template['rename'] = object_key.replace(change_from, change_to)
        if additional_parameters:
            template.update(additional_parameters)
    return template


def build_exp_type_query(exp_type, kwargs):
    assert exp_type in accepted_versions
    statuses = ['pre-release', 'released', 'released to project']
    versions = accepted_versions[exp_type]
    # Build the query
    pre_query = "/search/?experimentset_type=replicate&type=ExperimentSetReplicate"
    pre_query += "&experiments_in_set.experiment_type={}".format(exp_type)
    pre_query += "".join(["&status=" + i for i in statuses])
    # for some cases we don't have a defined complete processing tag
    if versions:
        pre_query += "".join(["&completed_processes!=" + i for i in versions])

    # skip non processable experiment sets
    pre_query += "&tags!=skip_processing"
    # add date
    s_date = kwargs.get('start_date')
    if s_date:
        pre_query += '&date_created.from=' + s_date
    # add lab
    lab = kwargs.get('lab_title')
    if lab:
        pre_query += '&lab.display_title=' + lab
    return pre_query


def build_feature_calling_query(exp_types, feature, kwargs):
    assert feature in feature_calling_accepted_versions

    for exp_type in exp_types:
        assert exp_type in accepted_versions

    statuses = ['pre-release', 'released', 'released to project', 'uploaded']
    versions = [i for i in accepted_versions[exp_type]]
    feature_calling_versions = feature_calling_accepted_versions[feature]
    # Build the query
    pre_query = "/search/?experimentset_type=replicate&type=ExperimentSetReplicate"
    pre_query += "".join(["&experiments_in_set.experiment_type=" + i for i in exp_types])
    pre_query += "".join(["&status=" + i for i in statuses])
    # for some cases we don't have a defined complete processing tag
    if versions:
        pre_query += "".join(["&completed_processes=" + i for i in versions])

    if feature_calling_versions:
        pre_query += "".join(["&completed_processes!=" + i for i in feature_calling_versions])
    pre_query += "&processed_files.quality_metric.@type=QualityMetricMcool"
    # add date
    s_date = kwargs.get('start_date')
    if s_date:
        pre_query += '&date_created.from=' + s_date
    # add lab
    lab = kwargs.get('lab_title')
    if lab:
        pre_query += '&lab.display_title=' + lab
    return pre_query


def find_fastq_info(my_rep_set, fastq_files, type=None):
    """Find fastq files from experiment set
    expects my_rep_set to be set response in frame object (search result)
    will check if files are paired or not, and if paired will give list of lists for each exp
    if not paired, with just give list of files per experiment.

    result is 2 dictionaries
    - file dict  { exp1 : [file1, file2, file3, file4]}  # unpaired
      file dict  { exp1 : [ [file1, file2], [file3, file4]]} # paired
    - refs keys  {pairing, organism, enzyme, bwa_ref, chrsize_ref, enz_ref, f_size, lab}
    """
    # remove non fastq.gz files from the file list
    fastq_files = [i for i in fastq_files if i['file_format']['file_format'] == 'fastq']
    file_dict = {}
    refs = {}

    # check pairing for the first file, and assume all same

    rep_resp = my_rep_set['experiments_in_set']
    enzymes = []
    organisms = []
    total_f_size = 0
    pairing = []  # collect pairing for each experiment and report if they are consistent or not
    for exp in rep_resp:
        paired = ""
        exp_resp = exp
        file_dict[exp['accession']] = []
        if not organisms:
            biosample = exp['biosample']
            organisms = list(set([bs.get('organism', {}).get('name') for bs in biosample['biosource']]))
        exp_files = exp['files']
        enzyme = exp.get('digestion_enzyme')
        if enzyme:
            enzymes.append(enzyme['display_title'])
        for fastq_file in exp_files:
            file_resp = [i for i in fastq_files if i['uuid'] == fastq_file['uuid']][0]
            if file_resp.get('file_size'):
                total_f_size += file_resp['file_size']
            # skip pair no 2
            if file_resp.get('paired_end') == '2':
                continue
            # check that file has a pair
            f1 = file_resp['@id']
            f2 = ""
            # assign pairing info by the first file
            if not paired:
                try:
                    relations = file_resp['related_files']
                    paired_files = [relation['file']['@id'] for relation in relations
                                    if relation['relationship_type'] == 'paired with']
                    assert len(paired_files) == 1
                    paired = "Yes"
                except:
                    paired = "No"

            if paired == 'No':
                file_dict[exp_resp['accession']].append(f1)
            elif paired == 'Yes':
                #
                relations = file_resp['related_files']
                paired_files = [relation['file']['@id'] for relation in relations
                                if relation['relationship_type'] == 'paired with']
                assert len(paired_files) == 1
                f2 = paired_files[0]
                file_dict[exp_resp['accession']].append((f1, f2))
        pairing.append(paired)
    # get the organism
    if len(list(set(organisms))) == 1:
        organism = organisms[0]
    elif len(list(set(organisms))) > 1:
        organism = "multiple organisms"
    else:
        organism = None

    # get the enzyme
    if len(list(set(enzymes))) == 1:
        enz = enzymes[0]
    else:
        enz = ''

    bwa = bwa_index.get(organism)
    chrsize = chr_size.get(organism)

    # if margi, enzyme files are predifined in a separate dict
    if type == 'MARGI':
        if re_fragment.get(organism):
            enz_file = re_fragment[organism].get(enz)
        else:
            enz_file = None
    else:
        # get the enzyme file for organism and enzyme type
        if re_nz.get(organism):
            enz_file = re_nz[organism].get(enz)
        else:
            enz_file = None

    f_size = int(total_f_size / (1024 * 1024 * 1024))
    # check pairing consistency
    if len(set(pairing)) == 1:
        set_pair_status = pairing[0]
    else:
        set_pair_status = 'Inconsistent'
    refs = {'pairing': set_pair_status,
            'organism': organism,
            'enzyme': enz,
            'bwa_ref': bwa,
            'chrsize_ref': chrsize,
            'enz_ref': enz_file,
            'f_size': str(f_size)+'GB'}
    return file_dict, refs


def check_runs_without_output(res, check, run_name, my_auth, start):
    """Common processing for checks that are running on files and not producing output files
    like qcs ones producing extra files"""
    # no successful run
    missing_run = []
    # successful run but no expected metadata change (qc or extra file)
    missing_meta_changes = []
    # still running
    running = []
    # multiple failed runs
    problems = []

    for a_file in res:
        # lambda has a time limit (300sec), kill before it is reached so we get some results
        now = datetime.utcnow()
        if (now-start).seconds > lambda_limit:
            check.brief_output.append('did not complete checking all')
            break
        file_id = a_file['accession']
        report = get_wfr_out(a_file, run_name,  key=my_auth, md_qc=True)
        if report['status'] == 'running':
            running.append(file_id)
        elif report['status'].startswith("no complete run, too many"):
            problems.append(file_id)
        elif report['status'] != 'complete':
            missing_run.append(file_id)
        # There is a successful run, but not the expected change (should be part of query)
        elif report['status'] == 'complete':
            missing_meta_changes.append(file_id)
    if running:
        check.summary = 'Some files are running'
        check.brief_output.append(str(len(running)) + ' files are still running.')
        check.full_output['running'] = running
    if problems:
        check.summary = 'Some files have problems'
        check.brief_output.append(str(len(problems)) + ' files have multiple failed runs')
        check.full_output['problems'] = problems
        check.status = 'WARN'
    if missing_run:
        check.allow_action = True
        check.summary = 'Some files are missing runs'
        check.brief_output.append(str(len(missing_run)) + ' files lack a successful run')
        check.full_output['files_without_run'] = missing_run
        check.status = 'WARN'
    if missing_meta_changes:
        check.allow_action = True
        check.summary = 'Some files are missing runs'
        check.brief_output.append(str(len(missing_meta_changes)) + ' files have successful run but no qc/extra file')
        check.full_output['files_without_changes'] = missing_meta_changes
        check.status = 'WARN'
    check.summary = check.summary.strip()
    if not check.brief_output:
        check.brief_output = ['All Good!']
    return check


def check_hic(res, my_auth, tag, check, start, lambda_limit, nore=False, nonorm=False):
    """Check run status for each set in res, and report missing runs and completed process"""
    for a_set in res:
        # get all related items
        all_items, all_uuids = ff_utils.expand_es_metadata([a_set['uuid']], my_auth,
                                                           store_frame='embedded',
                                                           add_pc_wfr=True,
                                                           ignore_field=['experiment_relation',
                                                                         'biosample_relation',
                                                                         'references',
                                                                         'reference_pubs'])
        all_wfrs = all_items.get('workflow_run_awsem', []) + all_items.get('workflow_run_sbg', [])
        now = datetime.utcnow()
        print(a_set['accession'], (now-start).seconds)
        if (now-start).seconds > lambda_limit:
            break
        # missing run
        missing_run = []
        # still running
        running = []
        # problematic cases
        problematic_run = []
        # if all runs are complete, add the patch info for processed files and tag
        complete = {'patch_opf': [],
                    'add_tag': []}
        set_summary = ""
        set_acc = a_set['accession']
        part3 = 'ready'
        # references dict content
        # pairing, organism, enzyme, bwa_ref, chrsize_ref, enz_ref, f_size
        exp_files, refs = find_fastq_info(a_set, all_items['file_fastq'])
        set_summary = " - ".join([set_acc, str(refs['organism']), str(refs['enzyme']), str(refs['f_size'])])
        # if no files were found
        if all(not value for value in exp_files.values()):
            set_summary += "| skipped - no usable file"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no usable file'})
            continue

        # Skip is organism is missing
        if not refs['organism']:
            set_summary += "| skipped - no organism"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no organism'})
            continue
        # skip if more than one organism
        if refs['organism'] == "multiple organisms":
            set_summary += "| skipped - multiple organisms"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - multiple organisms'})
            continue

        # skip if missing reference
        if not refs['bwa_ref'] or not refs['chrsize_ref']:
            set_summary += "| skipped - no chrsize/bwa"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no chrsize/bwa'})
            continue
        # if enzyme is used in the pipeline, check required parameters
        if not nore:
            if not refs['enzyme']:
                set_summary += "| skipped - missing enzyme metadata"
                check.brief_output.append(set_summary)
                check.full_output['skipped'].append({set_acc: 'skipped - missing enzyme metadata'})
                continue
            if not refs['enz_ref']:
                set_summary += "| skipped - no reference NZ file"
                check.brief_output.append(set_summary)
                check.full_output['skipped'].append({set_acc: 'skipped - no reference NZ file'})
                continue
        set_pairs = []
        # cycle through the experiments, skip the ones without usable files
        for exp in exp_files.keys():
            if not exp_files.get(exp):
                continue
            # Check Part 1 and See if all are okay
            exp_bams = []
            part2 = 'ready'
            for pair in exp_files[exp]:
                pair_resp = [i for i in all_items['file_fastq'] if i['@id'] == pair[0]][0]
                step1_result = get_wfr_out(pair_resp, 'bwa-mem', all_wfrs=all_wfrs)
                # if successful
                if step1_result['status'] == 'complete':
                    exp_bams.append(step1_result['out_bam'])
                # if still running
                elif step1_result['status'] == 'running':
                    part2 = 'not ready'
                    running.append(['step1', exp, pair])
                # if run is not successful
                elif step1_result['status'].startswith("no complete run, too many"):
                    part2 = 'not ready'
                    problematic_run.append(['step1', exp, pair])
                else:
                    part2 = 'not ready'
                    # add part 1
                    inp_f = {'fastq1': pair[0], 'fastq2': pair[1], 'bwa_index': refs['bwa_ref']}
                    name_tag = pair[0].split('/')[2]+'_'+pair[1].split('/')[2]
                    missing_run.append(['step1', ['bwa-mem', refs['organism'], {}], inp_f, name_tag])
            # stop progress to part2 and 3
            if part2 is not 'ready':
                part3 = 'not ready'
                # skip part 2 checks
                continue
            # make sure all input bams went through same last step2
            all_step2s = []
            for bam in exp_bams:
                bam_resp = [i for i in all_items['file_processed'] if i['@id'] == bam][0]
                step2_result = get_wfr_out(bam_resp, 'hi-c-processing-bam', all_wfrs=all_wfrs)
                all_step2s.append((step2_result['status'], step2_result.get('annotated_bam')))
            # all bams should have same wfr
            assert len(list(set(all_step2s))) == 1
            # check if part 2 run already
            if step2_result['status'] == 'complete':
                # accumulate pairs files for step3
                set_pairs.append(step2_result['filtered_pairs'])
                # add files for experiment opf
                patch_data = [step2_result['annotated_bam'], step2_result['filtered_pairs']]
                complete['patch_opf'].append([exp, patch_data])
                continue
            # if still running
            elif step2_result['status'] == 'running':
                part3 = 'not ready'
                running.append(['step2', exp])
                continue
            # problematic runs with repeated fails
            elif step2_result['status'].startswith("no complete run, too many"):
                part3 = 'not ready'
                problematic_run.append(['step2', exp])
                continue
            # if run is not successful
            else:
                part3 = 'not ready'
                # Add part2
                inp_f = {'input_bams': exp_bams, 'chromsize': refs['chrsize_ref']}
                missing_run.append(['step2', ['hi-c-processing-bam', refs['organism'], {}], inp_f, exp])
        if part3 is not 'ready':
            if running:
                set_summary += "| running step 1/2"
            elif missing_run:
                set_summary += "| missing step 1/2"
            elif problematic_run:
                set_summary += "| problem in step 1/2"

        if part3 is 'ready':
            # if we made it to this step, there should be files in set_pairs
            assert set_pairs
            # make sure all input bams went through same last step3
            all_step3s = []
            for a_pair in set_pairs:
                a_pair_resp = [i for i in all_items['file_processed'] if i['@id'] == a_pair][0]
                step3_result = get_wfr_out(a_pair_resp, 'hi-c-processing-pairs', all_wfrs=all_wfrs)
                all_step3s.append((step3_result['status'], step3_result.get('mcool')))
            # make sure existing step3s are matching
            if len(list(set(all_step3s))) == 1:
                # if successful
                if step3_result['status'] == 'complete':
                    set_summary += '| completed runs'
                    patch_data = [step3_result['merged_pairs'], step3_result['hic'], step3_result['mcool']]
                    complete['patch_opf'].append([set_acc, patch_data])
                    complete['add_tag'] = [set_acc, tag]
                # if still running
                elif step3_result['status'] == 'running':
                    running.append(['step3', set_acc])
                    set_summary += "| running step3"
                # problematic runs with repeated fails
                elif step3_result['status'].startswith("no complete run, too many"):
                    set_summary += "| problems in step3"
                    problematic_run.append(['step3', set_acc])
                # if run is not successful
                else:
                    set_summary += "| missing step3"
                    inp_f = {'input_pairs': set_pairs,
                             'chromsizes': refs['chrsize_ref']}
                    if not nore:
                        inp_f['restriction_file'] = refs['enz_ref']
                    overwrite = {}
                    if nonorm:
                        overwrite = {'parameters': {"no_balance": True}}
                    missing_run.append(['step3', ['hi-c-processing-pairs', refs['organism'], overwrite],
                                        inp_f, set_acc])
            else:
                problematic_run.append(['step3-not_unique', set_acc, all_step3s])
                set_summary += "| problem in step 3- not unique"
        check.brief_output.append(set_summary)
        if running:
            check.full_output['running_runs'].append({set_acc: running})
        if missing_run:
            check.full_output['needs_runs'].append({set_acc: missing_run})
        if problematic_run:
            check.full_output['problematic_runs'].append({set_acc: problematic_run})
        # if made it till the end
        if complete.get('add_tag'):
            assert not running
            assert not problematic_run
            assert not missing_run
            check.full_output['completed_runs'].append(complete)
    # complete check values
    if check.full_output['running_runs']:
        check.summary = str(len(check.full_output['running_runs'])) + ' running|'
    if check.full_output['skipped']:
        check.summary += str(len(check.full_output['skipped'])) + ' skipped|'
        check.status = 'WARN'
    if check.full_output['needs_runs']:
        check.summary += str(len(check.full_output['needs_runs'])) + ' missing|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['completed_runs']:
        check.summary += str(len(check.full_output['completed_runs'])) + ' completed|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['problematic_runs']:
        check.summary += str(len(check.full_output['problematic_runs'])) + ' problem|'
        check.status = 'WARN'
    return check


def check_margi(res, my_auth, tag, check, start, lambda_limit, nore=False, nonorm=False):
    """Check run status for each set in res, and report missing runs and completed process"""
    for a_set in res:
        # get all related items
        all_items, all_uuids = ff_utils.expand_es_metadata([a_set['uuid']], my_auth,
                                                           store_frame='embedded',
                                                           add_pc_wfr=True,
                                                           ignore_field=['experiment_relation',
                                                                         'biosample_relation',
                                                                         'references',
                                                                         'reference_pubs'])
        all_wfrs = all_items.get('workflow_run_awsem', []) + all_items.get('workflow_run_sbg', [])
        now = datetime.utcnow()
        print(a_set['accession'], (now-start).seconds, len(all_uuids))
        if (now-start).seconds > lambda_limit:
            break
        # missing run
        missing_run = []
        # still running
        running = []
        # problematic cases
        problematic_run = []
        # if all runs are complete, add the patch info for processed files and tag
        complete = {'patch_opf': [],
                    'add_tag': []}
        set_summary = ""
        set_acc = a_set['accession']
        part3 = 'ready'
        # references dict content
        # pairing, organism, enzyme, bwa_ref, chrsize_ref, enz_ref, f_size
        exp_files, refs = find_fastq_info(a_set, all_items['file_fastq'], type='MARGI')
        set_summary = " - ".join([set_acc, str(refs['organism']), str(refs['enzyme']), str(refs['f_size'])])
        # if no files were found
        if all(not value for value in exp_files.values()):
            set_summary += "| skipped - no usable file"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no usable file'})
            continue
        # Skip is organism is missing
        if not refs['organism']:
            set_summary += "| skipped - no organism"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no organism'})
            continue
        # skip if more than one organism
        if refs['organism'] == "multiple organisms":
            set_summary += "| skipped - multiple organisms"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - multiple organisms'})
            continue
        # skip if missing reference
        if not refs['bwa_ref'] or not refs['chrsize_ref']:
            set_summary += "| skipped - no chrsize/bwa"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no chrsize/bwa'})
            continue
        if not refs['enz_ref'] and not nore:
            set_summary += "| skipped - no enz"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no enz'})
            continue
        if refs['pairing'] != 'Yes':
            set_summary += "| skipped - unpaired files"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - not paired'})
            continue

        set_pairs = []
        # cycle through the experiments, skip the ones without usable files
        for exp in exp_files.keys():
            # we don't have exp level runs, but we have a patch, hence part 2.5
            part2_5 = 'ready'
            if not exp_files.get(exp):
                continue
            # Check Part 1 and See if all are okay
            exp_pairs = []
            exp_margi_files = []
            for pair in exp_files[exp]:
                part2 = 'ready'
                input_bam = ""
                pair_resp = [i for i in all_items['file_fastq'] if i['@id'] == pair[0]][0]
                step1_result = get_wfr_out(pair_resp, 'imargi-processing-fastq', all_wfrs=all_wfrs)
                # if successful
                if step1_result['status'] == 'complete':
                    input_bam = step1_result['out_bam']
                    exp_margi_files.append(step1_result['out_bam'])
                # if still running
                elif step1_result['status'] == 'running':
                    part2 = 'not ready'
                    running.append(['step1', exp, pair])
                # if run is not successful
                elif step1_result['status'].startswith("no complete run, too many"):
                    part2 = 'not ready'
                    problematic_run.append(['step1', exp, pair])
                else:
                    part2 = 'not ready'
                    # add part 1
                    inp_f = {'fastq_R1': pair[0], 'fastq_R2': pair[1], 'bwa_index': refs['bwa_ref']}
                    name_tag = pair[0].split('/')[2]+'_'+pair[1].split('/')[2]
                    missing_run.append(['step1', ['imargi-processing-fastq', refs['organism'], {}], inp_f, name_tag])

                # going into step2
                if part2 != 'ready':
                    part2_5 = 'not ready'
                    continue

                bam_resp = [i for i in all_items['file_processed'] if i['@id'] == input_bam][0]
                step2_result = get_wfr_out(bam_resp, 'imargi-processing-bam', all_wfrs=all_wfrs)
                # if successful
                if step2_result['status'] == 'complete':
                    exp_margi_files.append(step2_result['out_pairs'])
                    exp_pairs.append(step2_result['out_pairs'])
                # if still running
                elif step2_result['status'] == 'running':
                    part2_5 = 'not ready'
                    running.append(['step2', exp, input_bam])
                # if run is not successful
                elif step2_result['status'].startswith("no complete run, too many"):
                    part2_5 = 'not ready'
                    problematic_run.append(['step2', exp, input_bam])
                else:
                    part2_5 = 'not ready'
                    # add part 1
                    inp_f = {'input_bam': input_bam, 'chromsize': refs['chrsize_ref'],
                             'restrict_frags': refs['enz_ref']}
                    name_tag = input_bam.split('/')[2]
                    missing_run.append(['step2', ['imargi-processing-bam', refs['organism'], {}], inp_f, name_tag])

            if part2_5 != 'ready':
                part3 = 'not ready'
            else:
                # if exps runs were fine, lets patch exp with all pairs produced
                patch_data = exp_margi_files
                complete['patch_opf'].append([exp, patch_data])
                set_pairs.extend(exp_pairs)

                # patch the experiment with exp_pairs
        if part3 is not 'ready':
            if running:
                set_summary += "| running step 1/2"
            elif missing_run:
                set_summary += "| missing step 1/2"
            elif problematic_run:
                set_summary += "| problem in step 1/2"

        if part3 is 'ready':
            # if we made it to this step, there should be files in set_pairs
            assert set_pairs
            # make sure all input bams went through same last step3
            all_step3s = []
            for a_pair in set_pairs:
                a_pair_resp = [i for i in all_items['file_processed'] if i['@id'] == a_pair][0]
                step3_result = get_wfr_out(a_pair_resp, 'imargi-processing-pairs', all_wfrs=all_wfrs)
                all_step3s.append((step3_result['status'], step3_result.get('out_mcool')))
            # make sure existing step3s are matching
            if len(list(set(all_step3s))) == 1:
                # if successful
                if step3_result['status'] == 'complete':
                    set_summary += '| completed runs'
                    patch_data = [step3_result['merged_pairs'], step3_result['out_mcool']]
                    complete['patch_opf'].append([set_acc, patch_data])
                    complete['add_tag'] = [set_acc, tag]
                # if still running
                elif step3_result['status'] == 'running':
                    running.append(['step3', set_acc])
                    set_summary += "| running step3"
                # problematic runs with repeated fails
                elif step3_result['status'].startswith("no complete run, too many"):
                    set_summary += "| problems in step3"
                    problematic_run.append(['step3', set_acc])
                # if run is not successful
                else:
                    set_summary += "| missing step3"
                    inp_f = {'input_pairs': set_pairs}
                    missing_run.append(['step3', ['imargi-processing-pairs', refs['organism'], {}],
                                        inp_f, set_acc])
            else:
                problematic_run.append(['step3-not_unique', set_acc])
                set_summary += "| problem in step 3- not unique"
        check.brief_output.append(set_summary)
        if running:
            check.full_output['running_runs'].append({set_acc: running})
        if missing_run:
            check.full_output['needs_runs'].append({set_acc: missing_run})
        if problematic_run:
            check.full_output['problematic_runs'].append({set_acc: problematic_run})
        # if made it till the end
        if complete.get('add_tag'):
            assert not running
            assert not problematic_run
            assert not missing_run
            check.full_output['completed_runs'].append(complete)
    # complete check values
    if check.full_output['running_runs']:
        check.summary = str(len(check.full_output['running_runs'])) + ' running|'
    if check.full_output['skipped']:
        check.summary += str(len(check.full_output['skipped'])) + ' skipped|'
        check.status = 'WARN'
    if check.full_output['needs_runs']:
        check.summary += str(len(check.full_output['needs_runs'])) + ' missing|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['completed_runs']:
        check.summary += str(len(check.full_output['completed_runs'])) + ' completed|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['problematic_runs']:
        check.summary += str(len(check.full_output['problematic_runs'])) + ' problem|'
        check.status = 'WARN'
    return check


def patch_complete_data(patch_data, pipeline_type, auth, move_to_pc=False, pc_append=False):
    """Function to update experiment and experiment set metadata for pipeline completions
    and output files.
    Parameters
    ----------
    patch_data: (dict) example format:
                {
                'patch_opf': [
                    ['set_acc', ['file1', 'file2']],
                    ['exp_acc', ['file3', 'file4']]
                    ]
                'add_tag': [['exp_acc', 'completed_pipeline_tag']]
                }
    pipeline_type: (str) key for titles dictionary for setting the opf title
    move_to_pc: (bool) If False, processing results go to other_processed_files field
                If True:
                   If set/exp is released/to project processing results go to other_processed_files field
                   If set/exp is in other status, processing results go to processed_files field
    pc_append: (bool) If True and move_to_pc is True - append outfiles to existing processed files
            This is relevant for pipelines that produce files to be added to datasets upon which a pipeline (eg. Hi-C)
            has already been run - eg. Compartment Caller or Insulation Score/Boundaries
    """
    titles = {"hic": "HiC Processing Pipeline - Preliminary Files",
              "repliseq": "Repli-Seq Pipeline - Preliminary Files",
              'chip': "ENCODE ChIP-Seq Pipeline - Preliminary Files",
              'atac': "ENCODE ATAC-Seq Pipeline - Preliminary Files",
              'margi': "iMARGI Processing Pipeline - Preliminary Files",
              'rnaseq': "ENCODE RNA-Seq Pipeline - Preliminary Files",
              'insulation_scores_and_boundaries': "Insulation scores and boundaries calls - Preliminary Files",
              'compartments': "Compartments Signals - Preliminary Files"}
    """move files to other processed_files field."""
    if not patch_data.get('patch_opf'):
        return ['no content in patch_opf, skipping']
    if not patch_data.get('add_tag'):
        return ['no tag info, skipping']
    pc_set_title = titles[pipeline_type]
    log = []
    for a_case in patch_data['patch_opf']:
        # exp/set acc, and list of files to add
        acc, list_pc = a_case[0], a_case[1]
        resp = ff_utils.get_metadata(acc, auth)
        # check if these items are in existing processed files field
        ex_pc = resp.get('processed_files')
        if ex_pc:
            ex_pc_ids = [i['@id'] for i in ex_pc]
            common = list(set(ex_pc_ids) & set(list_pc))
            if common:
                log.append('some files ({}) are already in processed_files filed for {}'.format(common, acc))
                continue
        # check if these items are in other processed files field
        ex_opc = resp.get('other_processed_files')
        if ex_opc:
            # make sure the title is not already There
            all_existing_titles = [a['title'] for a in ex_opc]
            if pc_set_title in all_existing_titles:
                log.append('opc using same title already exists for {}'.format(acc))
                continue
            ex_opc_ids = [i['@id'] for a in ex_opc for i in a['files']]
            common = list(set(ex_opc_ids) & set(list_pc))
            if common:
                log.append('some files ({}) are already in other_processed_files filed for {}'.format(common, acc))
                continue
        source_status = resp['status']
        # if move_to_pc is set to true, but the source status is released/to project
        # set it back to finalize_user_pending_labs
        if source_status in ['released', 'released to project']:
            move_to_pc = False
        # if move_to_pc is true, add them to processed_files
        if move_to_pc:
            # at this step we expect processed_files field to be empty
            # unless pc_append is True
            if ex_pc:
                if pc_append:
                    ex_pc_cnt = len(ex_pc_ids)
                    ex_pc_ids.extend(list_pc)
                    list_pc = ex_pc_ids
                    if ex_pc_cnt == len(list_pc):
                        # warn if it looks like no existing pfs were added
                        log.append('expected additions to existing processed files: {}'.format(acc))
                        continue
                else:
                    log.append('expected processed_files to be empty: {}'.format(acc))
                    continue
            # patch the processed files field
            ff_utils.patch_metadata({'processed_files': list_pc}, obj_id=acc, key=auth)
        # if not move_to_pc, add files to opf with proper title
        else:
            # we need raw to get the existing piece, to patch back with the new ones
            if ex_opc:
                patch_val = ff_utils.get_metadata(acc, key=auth, add_on='frame=raw').get('other_processed_files', [])
            else:
                patch_val = []

            new_data = {'title': pc_set_title,
                        'type': 'preliminary',
                        'files': list_pc}
            patch_val.append(new_data)
            patch_body = {'other_processed_files': patch_val}
            ff_utils.patch_metadata(patch_body, obj_id=acc, key=auth)
    # add the tag
    set_acc = patch_data['add_tag'][0]
    new_tag = patch_data['add_tag'][1]
    existing_tags = ff_utils.get_metadata(set_acc, auth).get('completed_processes', [])
    new_tags = list(set(existing_tags + [new_tag]))
    ff_utils.patch_metadata({'completed_processes': new_tags}, set_acc, auth)
    return log


def run_missing_wfr(input_json, input_files_and_params, run_name, auth, env, fs_env, mount=False):
    if fs_env == 'staging':
        raise ValueError("'staging' not an expected value for fs_env - pipelines do not run on staging."
                         "please run on data instead.")
    all_inputs = []
    # input_files container
    input_files = {k: v for k, v in input_files_and_params.items() if k != 'additional_file_parameters'}
    # additional input file parameters
    input_file_parameters = input_files_and_params.get('additional_file_parameters', {})
    for arg, files in input_files.items():
        additional_params = input_file_parameters.get(arg, {})
        inp = extract_file_info(files, arg, additional_params, auth, env)
        all_inputs.append(inp)
    # tweak to get bg2bw working
    all_inputs = sorted(all_inputs, key=itemgetter('workflow_argument_name'))
    my_s3_util = s3Utils(env=env)
    out_bucket = my_s3_util.outfile_bucket
    sfn = 'tibanna_pony_' + fs_env
    # shorten long name_tags
    # they get combined with workflow name, and total should be less then 80
    # (even less since repeats need unique names)
    if len(run_name) > 30:
        run_name = run_name[:30] + '...'
    """Creates the trigger json that is used by foufront endpoint.
    """
    input_json['input_files'] = all_inputs
    input_json['output_bucket'] = out_bucket
    input_json["_tibanna"] = {
        "env": env,
        "run_type": input_json['app_name'],
        "run_id": run_name}
    input_json['public_postrun_json'] = True
    input_json['step_function_name'] = sfn
    input_json['env_name'] = env
    if mount:
        for a_file in input_json['input_files']:
            a_file['mount'] = True

    # # testing
    # json_object = json.dumps(input_json, indent=4)
    # print(json_object)
    # return

    # env should be either data, webdev or fourfront-webdev

    try:
        res = API().run_workflow(input_json, sfn=sfn, verbose=False)
        url = res['_tibanna']['url']
        return url
    except Exception as e:
        return str(e)


def start_missing_run(run_info, auth, env, fs_env):
    attr_keys = ['fastq1', 'fastq', 'input_pairs', 'input_bams', 'input_fastqs',
                 'fastq_R1', 'input_bam', 'rna.fastqs_R1', 'mad_qc.quantfiles', 'mcoolfile',
                 'chip.ctl_fastqs', 'chip.fastqs', 'chip.tas', 'atac.fastqs', 'atac.tas']
    run_settings = run_info[1]
    inputs = run_info[2]
    name_tag = run_info[3]
    attr_file = ''
    # find file to use for attribution
    for attr_key in attr_keys:
        if attr_key in inputs:
            attr_file = inputs[attr_key]
            if isinstance(attr_file, list):
                attr_file = attr_file[0]
                if isinstance(attr_file, list):
                    attr_file = attr_file[0]
                    if isinstance(attr_file, list):
                        attr_file = attr_file[0]
                        break
                    else:
                        break
                else:
                    break
            else:
                break
    if not attr_file:
        possible_keys = [i for i in inputs.keys() if i != 'additional_file_parameters']
        error_message = ('one of these argument names {} which carry the input file -not the references-'
                         ' should be added to att_keys dictionary on foursight cgap_utils.py function start_missing_run').format(possible_keys)
        raise ValueError(error_message)
    attributions = get_attribution(ff_utils.get_metadata(attr_file, auth))
    settings = wfrset_utils.step_settings(run_settings[0], run_settings[1], attributions, run_settings[2])
    url = run_missing_wfr(settings, inputs, name_tag, auth, env, fs_env, mount=False)
    return url


def start_tasks(missing_runs, patch_meta, action, my_auth, my_env, fs_env, start, move_to_pc=False, runtype='hic', pc_append=False):
    started_runs = 0
    action.description = ""
    action_log = {'started_runs': [], 'failed_runs': [], 'patched_meta': [], 'failed_meta': []}
    if missing_runs:
        for a_case in missing_runs:
            now = datetime.utcnow()
            acc = list(a_case.keys())[0]
            print((now-start).seconds, acc)
            if (now-start).seconds > lambda_limit:
                action.description = 'Did not complete action due to time limitations.'
                break

            for a_run in a_case[acc]:
                started_runs += 1
                url = start_missing_run(a_run, my_auth, my_env, fs_env)
                log_message = acc + ' started running ' + a_run[0] + ' with ' + a_run[3]
                if url.startswith('http'):
                    action_log['started_runs'].append([log_message, url])
                else:
                    action_log['failed_runs'].append([log_message, url])
    if patch_meta:
        action_log['patched_meta'] = []
        for a_completed_info in patch_meta:
            acc = a_completed_info['add_tag'][0]
            now = datetime.utcnow()
            if (now-start).seconds > lambda_limit:
                action.description = 'Did not complete action due to time limitations.'
                break
            error = patch_complete_data(a_completed_info, runtype, my_auth, move_to_pc=move_to_pc, pc_append=pc_append)
            if not error:
                log_message = acc + ' completed processing'
                action_log['patched_meta'].append(log_message)
            else:
                action_log['failed_meta'].append([acc, error])

    # did we complete without running into time limit
    for k in action_log:
        if action_log[k]:
            add_desc = "| {}: {} ".format(k, str(len(action_log[k])))
            action.description += add_desc

    action.output = action_log
    action.status = 'DONE'
    return action


def check_repli(res, my_auth, tag, check, start, lambda_limit, winsize=None):
    """Check run status for each set in res, and report missing runs and completed process"""
    for a_set in res:
        # get all related items
        all_items, all_uuids = ff_utils.expand_es_metadata([a_set['uuid']], my_auth,
                                                           store_frame='embedded',
                                                           add_pc_wfr=True,
                                                           ignore_field=['experiment_relation',
                                                                         'biosample_relation',
                                                                         'references',
                                                                         'reference_pubs'])
        all_wfrs = all_items.get('workflow_run_awsem', []) + all_items.get('workflow_run_sbg', [])
        now = datetime.utcnow()
        print(a_set['accession'], (now-start).seconds)
        if (now-start).seconds > lambda_limit:
            break
        # missing run
        missing_run = []
        # still running
        running = []
        # problematic cases
        problematic_run = []
        # if all runs are complete, add the patch info for processed files and tag
        complete = {'patch_opf': [],
                    'add_tag': []}
        set_summary = ""
        set_acc = a_set['accession']
        # references dict content
        # pairing, organism, enzyme, bwa_ref, chrsize_ref, enz_ref, f_size
        exp_files, refs = find_fastq_info(a_set, all_items['file_fastq'])
        paired = refs['pairing']
        set_summary = " - ".join([set_acc, str(refs['organism']), str(refs['f_size'])])
        # if no files were found
        if all(not value for value in exp_files.values()):
            set_summary += "| skipped - no usable file"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no usable file'})
            continue
        # Skip is organism is missing
        if not refs['organism']:
            set_summary += "| skipped - no organism"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no organism'})
            continue
        # skip if more than one organism
        if refs['organism'] == "multiple organisms":
            set_summary += "| skipped - multiple organisms"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - multiple organisms'})
            continue
        # skip if missing reference
        if not refs['bwa_ref'] or not refs['chrsize_ref']:
            set_summary += "| skipped - no chrsize/bwa"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no chrsize/bwa'})
            continue
        # cycle through the experiments, skip the ones without usable files
        part3 = 'ready'  # switch for watching the set
        for exp in exp_files.keys():
            if not exp_files.get(exp):
                continue
            # Check Part 1 and See if all are okay
            all_files = []
            part2 = 'ready'  # switch for watching the exp
            for pair in exp_files[exp]:
                if paired == 'Yes':
                    pair_resp = [i for i in all_items['file_fastq'] if i['@id'] == pair[0]][0]
                elif paired == 'No':
                    pair_resp = [i for i in all_items['file_fastq'] if i['@id'] == pair][0]
                step1_result = get_wfr_out(pair_resp, 'repliseq-parta', all_wfrs=all_wfrs)
                # if successful
                if step1_result['status'] == 'complete':
                    all_files.extend([step1_result['filtered_sorted_deduped_bam'],
                                      step1_result['count_bg']])
                # if still running
                elif step1_result['status'] == 'running':
                    part2 = 'not ready'
                    running.append(['step1', exp, pair])
                # if run is not successful
                elif step1_result['status'].startswith("no complete run, too many"):
                    part2 = 'not ready'
                    problematic_run.append(['step1', exp, pair])
                else:
                    part2 = 'not ready'
                    # add part 1
                    if paired == 'Yes':
                        inp_f = {'fastq': pair[0], 'fastq2': pair[1],
                                 'bwaIndex': refs['bwa_ref'],
                                 'chromsizes': refs['chrsize_ref']}
                        name_tag = pair[0].split('/')[2]+'_'+pair[1].split('/')[2]
                    elif paired == 'No':
                        inp_f = {'fastq': pair,
                                 'bwaIndex': refs['bwa_ref'],
                                 'chromsizes': refs['chrsize_ref']}
                        name_tag = pair.split('/')[2]
                    overwrite = {}
                    if winsize:
                        overwrite = {'parameters': {'winsize': winsize}}
                    missing_run.append(['step1', ['repliseq-parta', refs['organism'], overwrite], inp_f, name_tag])
            # are all step1s complete
            if part2 == 'ready':
                # add files for experiment opf
                complete['patch_opf'].append([exp, all_files])
            else:
                part3 = 'not ready'
        if part3 == 'ready':
            # add the tag
            set_summary += "| completed runs"
            complete['add_tag'] = [set_acc, tag]
        else:
            if running:
                set_summary += "| running step 1"
            elif missing_run:
                set_summary += "| missing step 1"
            elif problematic_run:
                set_summary += "| problem in step 1"

        check.brief_output.append(set_summary)
        if running:
            check.full_output['running_runs'].append({set_acc: running})
        if missing_run:
            check.full_output['needs_runs'].append({set_acc: missing_run})
        if problematic_run:
            check.full_output['problematic_runs'].append({set_acc: problematic_run})
        # if made it till the end
        if complete.get('add_tag'):
            assert not running
            assert not problematic_run
            assert not missing_run
            check.full_output['completed_runs'].append(complete)
    # complete check values
    check.summary = ""
    if check.full_output['running_runs']:
        check.summary = str(len(check.full_output['running_runs'])) + ' running|'
    if check.full_output['skipped']:
        check.summary += str(len(check.full_output['skipped'])) + ' skipped|'
        check.status = 'WARN'
    if check.full_output['needs_runs']:
        check.summary += str(len(check.full_output['needs_runs'])) + ' missing|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['completed_runs']:
        check.summary += str(len(check.full_output['completed_runs'])) + ' completed|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['problematic_runs']:
        check.summary += str(len(check.full_output['problematic_runs'])) + ' problem|'
        check.status = 'WARN'
    return check


def check_rna(res, my_auth, tag, check, start, lambda_limit):
    """Check run status for each set in res, and report missing runs and completed process"""
    for a_set in res:
        # get all related items
        all_items, all_uuids = ff_utils.expand_es_metadata([a_set['uuid']], my_auth,
                                                           store_frame='embedded',
                                                           add_pc_wfr=True,
                                                           ignore_field=['experiment_relation',
                                                                         'biosample_relation',
                                                                         'references',
                                                                         'reference_pubs'])
        all_wfrs = all_items.get('workflow_run_awsem', []) + all_items.get('workflow_run_sbg', [])
        now = datetime.utcnow()
        # print(a_set['accession'], (now-start).seconds)
        if (now-start).seconds > lambda_limit:
            break
        # missing run
        missing_run = []
        # still running
        running = []
        # problematic cases
        problematic_run = []
        # if all runs are complete, add the patch info for processed files and tag
        complete = {'patch_opf': [],
                    'add_tag': []}
        set_summary = ""
        set_acc = a_set['accession']
        final_status = 'ready'
        # references dict content
        # pairing, organism, enzyme, bwa_ref, chrsize_ref, enz_ref, f_size
        exp_files, refs = find_fastq_info(a_set, all_items['file_fastq'])

        print(a_set['accession'], 'paired=', refs['pairing'], refs['organism'], refs['f_size'])
        paired = refs['pairing']
        organism = refs['organism']
        set_summary = " - ".join([set_acc, str(organism), str(refs['f_size'])])

        # if no files were found
        if all(not value for value in exp_files.values()):
            set_summary += "| skipped - no usable file"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no usable file'})
            continue
        # Skip is organism is missing
        if not refs['organism']:
            set_summary += "| skipped - no organism"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - no organism'})
            continue
        # skip if more than one organism
        if refs['organism'] == "multiple organisms":
            set_summary += "| skipped - multiple organisms"
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: 'skipped - multiple organisms'})
            continue
        if organism not in ['mouse', 'human']:
            msg = 'No reference file for ' + organism
            set_summary += "| " + msg
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: msg})
            continue

        # check  strandedness_verified
        not_verified = []
        for an_exp in exp_files:
            an_exp_resp = [i for i in all_items['experiment_seq'] if i['accession'] == an_exp][0]
            tags = an_exp_resp.get('tags', [])
            if 'strandedness_verified' not in tags:
                not_verified.append(an_exp)
            elif not an_exp_resp.get('strandedness'):
                not_verified.append(an_exp)
        if not_verified:
            msg = ', '.join(not_verified) + ' Not verified for strandedness'
            set_summary += "| " + msg
            check.brief_output.append(set_summary)
            check.full_output['skipped'].append({set_acc: msg})
            continue

        # cycle through the experiments, skip the ones without usable files
        # accumulate files for madqc
        step2_files = []
        step2_status = 'ready'
        for exp in exp_files.keys():
            step1_status = 'ready'
            if not exp_files.get(exp):
                continue

            strand_info = ''
            exp_resp = [i for i in all_items['experiment_seq'] if i['accession'] == exp][0]
            tags = exp_resp.get('tags', [])
            strand_info = exp_resp.get('strandedness')

            # run unstranded pipeline
            app_name = ''
            # set parameters
            pars = {
                'rna.strandedness': '',
                'rna.strandedness_direction': '',
                'rna.endedness': ''
            }
            if strand_info == 'unstranded':
                pars['rna.strandedness'] = 'unstranded'
                pars['rna.strandedness_direction'] = 'unstranded'
                app_name = 'encode-rnaseq-unstranded'
            elif strand_info in ['reverse', 'forward']:
                pars['rna.strandedness'] = 'stranded'
                pars['rna.strandedness_direction'] = strand_info
                app_name = 'encode-rnaseq-stranded'

            # add more parameters and get status report
            input_files = exp_files[exp]
            if paired == 'Yes':
                pars['rna.endedness'] = 'paired'
                input_resp = [i for i in all_items['file_fastq'] if i['@id'] == input_files[0][0]][0]
            elif paired == 'No':
                pars['rna.endedness'] = 'single'
                input_resp = [i for i in all_items['file_fastq'] if i['@id'] == input_files[0]][0]
            step1_result = get_wfr_out(input_resp, app_name, all_wfrs=all_wfrs)

            # if successful
            if step1_result['status'] == 'complete':
                # add  madqc file
                step2_files.append(step1_result['rna.gene_expression'])
                # create processed files list for experiment
                exp_results = []
                for a_type in ['rna.outbam',
                               'rna.plusbw',
                               'rna.minusbw',
                               'rna.outbw',
                               'rna.gene_expression',
                               'rna.isoform_expression']:
                    if a_type in step1_result:
                        exp_results.append(step1_result[a_type])
                complete['patch_opf'].append([exp, exp_results])
            # if still running
            elif step1_result['status'] == 'running':
                step1_status = 'not ready'
                running.append(['step1', exp])
            # if run is not successful
            elif step1_result['status'].startswith("no complete run, too many"):
                step1_status = 'not ready'
                problematic_run.append(['step1', exp])
            # if it is missing
            else:
                step1_status = 'not ready'
                # add part
                name_tag = exp
                inp_f = {'rna.align_index': rna_star_index[organism],
                         'rna.rsem_index': rna_rsem_index[organism],
                         'rna.chrom_sizes': rna_chr_size[organism],
                         'rna.rna_qc_tr_id_to_gene_type_tsv': rna_t2g[organism]}
                if paired == 'Yes':
                    inp_f['rna.fastqs_R1'] = [[i[0] for i in input_files]]
                    inp_f['rna.fastqs_R2'] = [[i[1] for i in input_files]]
                elif paired == 'No':
                    inp_f['rna.fastqs_R1'] = [input_files]
                overwrite = {'parameters': pars}
                missing_run.append(['step1', [app_name, organism, overwrite], inp_f, name_tag])
            if step1_status != 'ready':
                step2_status = 'not ready'

        if step2_status != 'ready':
            if running:
                set_summary += "| running step 1"
            elif missing_run:
                set_summary += "| missing step 1"
            elif problematic_run:
                set_summary += "| problem in step 1"
        # if there is a single replicate, skip madqc
        elif len(step2_files) == 1:
            step2_status = 'ready'
        # run step2 if step1 s are complete
        else:
            step2_input = [i for i in all_items['file_processed'] if i['@id'] == step2_files[0]][0]
            step2_result = get_wfr_out(step2_input, 'mad_qc_workflow', all_wfrs=all_wfrs, md_qc=True)

            # if successful
            if step2_result['status'] == 'complete':
                pass
            # if still running
            elif step2_result['status'] == 'running':
                step2_status = 'not ready'
                running.append(['step2', set_acc])
            # if run is not successful
            elif step2_result['status'].startswith("no complete run, too many"):
                step2_status = 'not ready'
                problematic_run.append(['step2', set_acc])
            # if it is missing
            else:
                step2_status = 'not ready'
                # add part
                name_tag = set_acc
                inp_f = {'mad_qc.quantfiles': step2_files}
                missing_run.append(['step2', ['mad_qc_workflow', organism, {}], inp_f, name_tag])
        if step2_status != 'ready':
            final_status = 'not ready'

        if final_status == 'ready':
            # add the tag
            set_summary += "| completed runs"
            complete['add_tag'] = [set_acc, tag]
        else:
            if running:
                set_summary += "| running step 2"
            elif missing_run:
                set_summary += "| missing step 2"
            elif problematic_run:
                set_summary += "| problem in step 2"

        check.brief_output.append(set_summary)
        if running:
            check.full_output['running_runs'].append({set_acc: running})
        if missing_run:
            check.full_output['needs_runs'].append({set_acc: missing_run})
        if problematic_run:
            check.full_output['problematic_runs'].append({set_acc: problematic_run})
        # if made it till the end
        if complete.get('add_tag'):
            assert not running
            assert not problematic_run
            assert not missing_run
            check.full_output['completed_runs'].append(complete)
    # complete check values
    check.summary = ""
    if check.full_output['running_runs']:
        check.summary = str(len(check.full_output['running_runs'])) + ' running|'
    if check.full_output['skipped']:
        check.summary += str(len(check.full_output['skipped'])) + ' skipped|'
        check.status = 'WARN'
    if check.full_output['needs_runs']:
        check.summary += str(len(check.full_output['needs_runs'])) + ' missing|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['completed_runs']:
        check.summary += str(len(check.full_output['completed_runs'])) + ' completed|'
        check.status = 'WARN'
        check.allow_action = True
    if check.full_output['problematic_runs']:
        check.summary += str(len(check.full_output['problematic_runs'])) + ' problem|'
        check.status = 'WARN'
    return check


def string_to_list(string):
    "Given a string that is either comma separated values, or a python list, parse to list"
    for a_sep in "'\":[] ":
        values = string.replace(a_sep, ",")
    values = [i.strip() for i in values.split(',') if i]
    return values


def fetch_wfr_associated(wfr_info):
    """Given wfr embedded frame, find associated output files and qcs"""
    wfr_as_list = []
    wfr_as_list.append(wfr_info['uuid'])
    if wfr_info.get('output_files'):
        for o in wfr_info['output_files']:
            if o.get('value'):
                wfr_as_list.append(o['value']['uuid'])
            elif o.get('value_qc'):
                wfr_as_list.append(o['value_qc']['uuid'])
    if wfr_info.get('output_quality_metrics'):
        for qc in wfr_info['output_quality_metrics']:
            if qc.get('value'):
                wfr_as_list.append(qc['value']['uuid'])
    if wfr_info.get('quality_metric'):
        wfr_as_list.append(wfr_info['quality_metric']['uuid'])
    return list(set(wfr_as_list))


def get_chip_info(f_exp_resp, all_items):
    """Gether the following information from the first experiment in the chip set"""
    control = ""  # True or False (True if set in scope is control)
    control_set = ""  # None (if no control exp is set), or the control experiment for the one in scope
    target_type = ""  # Histone or TF (or None for control)
    # get target
    targets = f_exp_resp.get('targeted_factor', [])
    # TODO: tag all control antibodies as control and make use of it here
    if targets:
        # use the tag from the first target, this assumes the rest follows the first one
        target = targets[0]
        target_info = [i for i in all_items['bio_feature'] if i['uuid'] == target['uuid']][0]
        # set to tf default and switch to histone if tagged so
        target_tags = target_info.get('tags', [])
        if not target_tags:
            target_type = None
        elif 'histone' in target_tags:
            target_type = 'histone'
        else:
            target_type = 'tf'
    else:
        target_type = None

    # get organism
    biosample = f_exp_resp['biosample']
    organism = list(set([bs['organism']['name'] for bs in biosample['biosource']]))[0]

    # get control information
    exp_relation = f_exp_resp.get('experiment_relation')
    if exp_relation:
        rel_type = [i['relationship_type'] for i in exp_relation]
        if 'control for' in rel_type:
            control = True
        if 'controlled by' in rel_type:
            control = False
            controls = [i['experiment'] for i in exp_relation if i['relationship_type'] == 'controlled by']
            if len(controls) != 1:
                print('multiple control experiments')
            else:
                cont_exp_resp = [i for i in all_items['experiment_seq'] if i['uuid'] == controls[0]['uuid']][0]
                cont_exp_info = cont_exp_resp['experiment_sets']
                control_set = [i['accession'] for i in cont_exp_info if i['@id'].startswith('/experiment-set-replicates/')][0]
    else:
        # if no relation is present
        # set it as if control when the target is None
        if not target_type:
            control = True
        # if there is target, but no relation, treat it as an experiment without control
        else:
            control = False
            control_set = None
    return control, control_set, target_type, organism


def get_chip_files(exp_resp, all_files):
    files = []
    paired = ""
    exp_files = exp_resp['files']
    for a_file in exp_files:
        f_t = []
        file_resp = [i for i in all_files if i['uuid'] == a_file['uuid']][0]
        # get pair end no
        pair_end = file_resp.get('paired_end')
        if pair_end == '2':
            paired = 'paired'
            continue
        # get paired file
        paired_with = ""
        relations = file_resp.get('related_files')
        if not relations:
            pass
        else:
            for relation in relations:
                if relation['relationship_type'] == 'paired with':
                    paired = 'paired'
                    paired_with = relation['file']['@id']
        # decide if data is not paired end reads
        if not paired_with:
            if not paired:
                paired = 'single'
            else:
                if paired != 'single':
                    print('inconsistent fastq pair info')
                    continue
            f_t.append(file_resp['@id'])
        else:
            f2 = [i for i in all_files if i['@id'] == paired_with][0]
            f_t.append(file_resp['@id'])
            f_t.append(f2['@id'])
        files.append(f_t)
    return files, paired


def select_best_2(file_list, all_files, all_qcs):
    scores = []
    # run it for list with at least 3 elements
    if len(file_list) < 3:
        return(file_list)

    for f in file_list:
        f_resp = [i for i in all_files if i['@id'] == f][0]
        qc = f_resp['quality_metric']
        qc_resp = [i for i in all_qcs if i['uuid'] == qc['uuid']][0]
        try:
            score = qc_resp['nodup_flagstat_qc'][0]['mapped']
        except Exception:
            score = qc_resp['ctl_nodup_flagstat_qc'][0]['mapped']
        scores.append((score, f))
    scores = sorted(scores, key=lambda x: -x[0])
    return [scores[0][1], scores[1][1]]


def limit_number_of_runs(check, my_auth):
    """Checks the number of workflow runs started in the past 6h. Return the
    number of remaining runs before hitting the rate limit of pulls from Docker
    Hub. This is currently 200 every 6 hours."""
    n_runs_max = 180  # this is set below the actual limit, since only a few workflows check this before running
    statuses = ['in review by lab', 'deleted', 'pre-release', 'released to project', 'released']
    query = '/search/?type=WorkflowRunAwsem' + ''.join(['&status=' + s for s in statuses])
    six_h_ago = datetime.now(timezone.utc) - timedelta(hours=6)
    query += '&date_created.from=' + datetime.strftime(six_h_ago, "%Y-%m-%d %H:%M")
    recent_runs = ff_utils.search_metadata(query, key=my_auth)
    n_runs_available = max(n_runs_max - len(recent_runs), 0)  # no negative results
    if n_runs_available == 0:
        check.status = 'PASS'
        check.brief_output = ['Waiting (max 6h) due to Docker Hub rate limit']
        check.summary = f'Limiting the number of workflow runs to {n_runs_max} every 6h'
        check.full_output = {}
    return check, n_runs_available



import time
import boto3  # PyCharm says nothing is arranging to load this - is that really so?
import shutil
import datetime
import tempfile

from botocore.exceptions import ClientError
from git import Repo
from dcicutils.ff_utils import get_metadata
from dcicutils.deployment_utils import EBDeployer
from dcicutils.beanstalk_utils import compute_ff_stg_env
from dcicutils.env_utils import is_fourfront_env, indexer_env_for_env
from dcicutils.misc_utils import ignored
from dcicutils.beanstalk_utils import compute_ff_prd_env, beanstalk_info

# Use confchecks to import decorators object and its methods for each check module
# rather than importing check_function, action_function, CheckResult, ActionResult
# individually - they're now part of class Decorators in foursight-core::decorators
# that requires initialization with foursight prefix.
from .helpers.confchecks import *


def try_to_describe_indexer_env(env):
    """
    Small helper that wraps beanstalk_info so we can easily ignore error for non-existent env.
    """
    # XXX: Should make beanstalk_info take something like if_does_not_exist= that can be 'error' or None,
    #      defaulting to 'error' so it's an incompatible change, but so we can pass None for our need.
    try:
        return beanstalk_info(env)
    except (IndexError, ClientError):
        return None  # Previously got IndexError if env does not exist, but now get a botocore.exceptions.ClientError
    except Exception:
        raise  # something else happened we should (probably) raise


def clone_repo_to_temporary_dir(repo='https://github.com/4dn-dcic/fourfront.git', name='fourfront'):
    """ Clones the given repo (default fourfront) to a temporary directory whose
        absolute path is returned.
        XXX: This code should be refactored into a contextmanager (probably in dcicutils).
             See PR292 comments.
    """
    tempdir = tempfile.mkdtemp(prefix=name)
    Repo.clone_from(url=repo, to_path=tempdir)
    return tempdir


def cleanup_tempdir(tempdir):
    """ Wrapper for shutil.rmtree that cleans up the (repo dir) we just created """
    shutil.rmtree(tempdir)


def are_counts_even(env):
    """ From dcicutils - to be ported over in this form at some point. """
    try:
        totals = get_metadata('/counts', ff_env=env)['db_es_total'].split()
    except Exception:
        # if we cant get counts page assume its False
        return False, {}

    # example value of split totals: ["DB:", "74048", "ES:", "74048"]
    db_total = totals[1]
    es_total = totals[3]
    if int(db_total) > int(es_total):
        return False, totals
    return True, totals


@check_function(env='fourfront-hotseat')
def indexer_server_status(connection, **kwargs):
    """ Checks the status of index servers (if any are up) """
    check = CheckResult(connection, 'indexer_server_status')
    check.action = 'terminate_indexer_server'
    env = kwargs.get('env')
    indexer_env = indexer_env_for_env(env)
    description = try_to_describe_indexer_env(indexer_env)  # verify an indexer is online
    if description is None:
        check.status = 'PASS'  # could have been terminated
        check.summary = 'Did not locate corresponding indexer env: ' \
                        'given env %s does not have indexer %s' % (env, indexer_env)
        check.allow_action = False
        return check

    try:
        indexing_finished, counts = are_counts_even(env)
    except Exception as e:  # XXX: This should be handled in dcicutils -Will 5/18/2020
        check.status = 'ERROR'
        check.summary = 'Failed to get indexing status of given env: %s' % str(e)
        check.allow_action = False
        return check

    if indexing_finished is True:
        check.status = 'PASS'
        check.summary = 'Env %s has finished indexing and is ready to be terminated' % env
        check.action_message = 'Safe to terminate indexer env: %s since indexing is finished' % env
        check.allow_action = True  # allow us to terminate
    else:
        check.status = 'WARN'
        check.allow_action = True
        check.action_message = 'Will terminate indexer env %s when indexing is not done yet!' % env
        check.summary = 'Env %s has not finished indexing with remaining counts %s' % (env, ' '.join(counts))

    check.full_output = indexer_env  # full_output contains indexer_env we checked
    return check


@action_function(forced_env=None)
def terminate_indexer_server(connection, **kwargs):
    """ Terminates the index_server stored in the full_output member of the associated check above """
    action = ActionResult(connection, 'terminate_indexer_server')
    related_indexer_server_status = action.get_associated_check_result(kwargs)
    forced_env = kwargs.get('forced_env', None)
    env_to_terminate = related_indexer_server_status.get('full_output', None) if forced_env is None else forced_env

    if env_to_terminate:
        client = boto3.client('elasticbeanstalk')
        success = EBDeployer.terminate_indexer_env(client, env_to_terminate)
        if success:
            action.status = 'DONE'
            action.output = 'Successfully triggered termination of indexer env %s' % env_to_terminate
        else:
            action.status = 'FAIL'
            action.output = 'Encountered an error while issuing terminate call, please check the AWS EB Console'
    else:
        action.status = 'FAIL'
        action.output = 'Did not get an env_to_terminate from the associated check - bad action call?'

    return action


@check_function(env='fourfront-hotseat', application_version=None)
def provision_indexer_environment(connection, **kwargs):
    """ Provisions an indexer environment for the given env. Note that only one indexer can be online
        per application (one for 4DN, one for CGAP).

        IMPORTANT: env is an EB ENV NAME ('data', 'staging' are NOT valid).
        IMPORTANT: This ecosystem will break some integrated tests in dcicutils
    """
    check = CheckResult(connection, 'provision_indexer_environment')
    env = kwargs.get('env', None)
    application_version = kwargs.get('application_version', None)

    if application_version is None:
        check.status = 'ERROR'
        check.summary = 'Did not provide application_version to deploy indexer, which is required. ' \
                        'Get this information from the EB Console.'
        return check

    def _deploy_indexer(e, version):
        indexer_env = indexer_env_for_env(e)
        description = try_to_describe_indexer_env(indexer_env)
        if description is not None:
            check.status = 'ERROR'
            check.summary = 'Tried to spin up indexer env for %s when one already exists for this portal' % e
            return False
        else:
            return EBDeployer.deploy_indexer(e, version)

    if is_fourfront_env(env):
        success = _deploy_indexer(env, application_version)
        if success:
            check.status = 'PASS'
            check.summary = 'Successfully triggered indexer-server provision for environment %s' % env
        else:
            check.status = 'ERROR'
            check.summary = 'An error occurred on deployment. Check the AWS EB Console.'
    else:
        check.status = 'FAIL'
        check.summary = 'Gave an unknown environment: %s' % env

    return check


def _deploy_application_to_beanstalk(connection, **kwargs):
    """ Creates and returns a CheckResult for 'deploy_application_to_beanstalk'.

        NOTE: this does NOT have the @check_function decorator because I'd like
        to "call this check" in other checks. The decorator will effectively cast
        the resulting check to a dictionary, breaking nested checks. This could be
        fixed by refactoring "store_formatted_result". - Will 05/28/2020
    """
    check = CheckResult(connection, 'deploy_application_to_beanstalk')
    env = kwargs.get('env', 'fourfront-mastertest')  # by default
    branch = kwargs.get('branch', 'master')  # by default deploy master
    application_version_name = kwargs.get('application_version_name', None)

    # error if we try to deploy prod
    if env == compute_ff_prd_env():
        check.status = 'ERROR'
        check.summary = 'Tried to deploy production env %s from Foursight, aborting.' % env
        return check

    if application_version_name is None:  # if not specified, use branch+timestamp
        application_version_name = 'foursight-package-%s-%s' % (branch, datetime.datetime.utcnow())

    repo_location = clone_repo_to_temporary_dir()

    try:
        packaging_was_successful = EBDeployer.build_application_version(repo_location, application_version_name,
                                                                        branch=branch)
        time.sleep(10)  # give EB some time to index the new template
        if packaging_was_successful:
            try:
                EBDeployer.deploy_new_version(env, repo_location, application_version_name)
                check.status = 'PASS'
                check.summary = 'Successfully deployed version %s to env %s' % (application_version_name, env)
            except Exception as e:
                check.status = 'ERROR'
                check.summary = 'Exception thrown while deploying: %s' % str(e)
        else:
            check.status = 'ERROR'
            check.summary = 'Could not package repository: %s' % packaging_was_successful
    except Exception as e:
        check.status = 'ERROR'
        check.summary = 'Exception thrown while building application version: %s' % str(e)
    finally:
        cleanup_tempdir(repo_location)

    return check


@check_function(env='fourfront-mastertest',
                branch='master',
                application_version_name=None, repo=None)
def deploy_application_to_beanstalk(connection, **kwargs):
    """ Deploys application to beanstalk under the given application_version name + a branch
        Uses the above helper method, returning the (decorated, thus stored) check directly.

        NOTE: CGAP requires kwargs!
    """
    return _deploy_application_to_beanstalk(connection, **kwargs)


def deploy_env(connection, env_to_deploy, application_name, check, **kwargs):
    """ For a given beanstalk environment, with a configured check json and given application name,
        deploys the environment, or returns an error.
    """
    this_check = CheckResult(connection, check)
    helper_check = _deploy_application_to_beanstalk(connection,
                                                    env=env_to_deploy,
                                                    branch='master',
                                                    **kwargs)
    if helper_check.status == 'PASS':
        this_check.status = 'PASS'
        this_check.summary = ('Successfully deployed {what} master to {where}'
                              .format(what=application_name, where=env_to_deploy))
    else:
        this_check.status = 'ERROR'
        this_check.summary = 'Error occurred during deployment, see full_output'
        this_check.full_output = helper_check.summary  # should have error message
    return this_check


@check_function()
def deploy_ff_staging(connection, **kwargs):
    """ Deploys Fourfront master to whoever staging is.
        Runs as part of the 'deployment_checks' schedule on data ONLY.
    """
    return deploy_env(
        connection,
        env_to_deploy=compute_ff_stg_env(),
        application_name="Fourfront",
        check='deploy_ff_staging',
        **kwargs)



import requests
import json
import datetime
import boto3
import time
import geocoder
from foursight_core.stage import Stage
from foursight_core.checks.helpers.sys_utils import (
    parse_datetime_to_utc,
    wipe_build_indices
)
from dcicutils import (
    ff_utils,
    es_utils,
    beanstalk_utils,
    env_utils
)

# Use confchecks to import decorators object and its methods for each check module
# rather than importing check_function, action_function, CheckResult, ActionResult
# individually - they're now part of class Decorators in foursight-core::decorators
# that requires initialization with foursight prefix.
from .helpers.confchecks import *


# XXX: put into utils?
FF_BLUE_ES_CLUSTER_DOMAIN = 'fourfront-blue-6-8'
FF_GREEN_ES_CLUSTER_DOMAIN = 'fourfront-green-6-8'
PROD_ES_CLUSTERS = [
    FF_BLUE_ES_CLUSTER_DOMAIN,
    FF_GREEN_ES_CLUSTER_DOMAIN
]
FF_TEST_CLUSTER = 'search-fourfront-testing-6-8-kncqa2za2r43563rkcmsvgn2fq.us-east-1.es.amazonaws.com:443'
TEST_ES_CLUSTERS = [
    FF_TEST_CLUSTER
]


@check_function()
def wipe_ff_build_indices(connection, **kwargs):
    """ Wipes build (number prefixed) indices (on fourfront-testing) """
    check = CheckResult(connection, 'wipe_ff_build_indices')
    return wipe_build_indices(FF_TEST_CLUSTER, check)


@check_function()
def elastic_search_space(connection, **kwargs):
    """ Checks that our ES nodes all have a certain amount of space remaining """
    check = CheckResult(connection, 'elastic_search_space')
    full_output = {}
    client = es_utils.create_es_client(connection.ff_es, True)
    # use cat.nodes to get id,diskAvail for all nodes, filter out empties
    node_space_entries = filter(None, [data.split() for data in client.cat.nodes(h='id,diskAvail').split('\n')])
    check.summary = check.description = None
    full_output['nodes'] = {}
    for _id, remaining_space in node_space_entries:
        if 'gb' not in remaining_space:
            if 'mb' not in remaining_space:
                check.status = 'FAIL'
                check.summary = check.description = 'At least one of the nodes in this env has no space remaining'
            else:
                check.status = 'WARN'
                check.summary = check.description = 'At least one of the nodes in this env is low on space'
        full_output['nodes'][_id.strip()] = { 'remaining_space': remaining_space }
    if check.summary is None:
        check.status = 'PASS'
        check.summary = check.description = 'All nodes have >1gb remaining disk space'
    check.full_output = full_output
    return check


def resolve_es_domain_name(connection):
    """ Resolves blue vs. green depending on the URL, throwing exception if something.  """
    if 'fourfront-blue' in connection.ff_es:
        return FF_BLUE_ES_CLUSTER_DOMAIN
    elif 'fourfront-green' in connection.ff_es:
        return FF_GREEN_ES_CLUSTER_DOMAIN
    else:
        raise Exception('Tried to run autoscaling check on non-production cluster! %s' % connection.ff_es)


@check_function()
def scale_down_elasticsearch_production(connection, **kwargs):
    """ Scales down Elasticsearch (production configuration).
        HOT (0600 to 2000 EST):
            Master:
                3x c5.large.elasticsearch
            Data:
                2x c5.2xlarge.elasticsearch
        COLD (2000 to 0600 EST):  This is what we are resizing to
            Master:
                None
            Data:
                3x c5.xlarge.elasticsearch
        XXX: should probably use constants in ElasticSearchServiceClient
        For now, must be explicitly triggered - but should be put on a schedule.
    """
    check = CheckResult(connection, 'scale_down_elasticsearch_production')
    es_client = es_utils.ElasticSearchServiceClient()
    success = es_client.resize_elasticsearch_cluster(
                domain_name=resolve_es_domain_name(connection),
                master_node_type='t2.medium.elasticsearch',  # discarded
                master_node_count=0,
                data_node_type='c5.xlarge.elasticsearch',
                data_node_count=3
            )
    if not success:
        check.status = 'ERROR'
        check.description = 'Could not trigger cluster resize - check lambda logs'
    else:
        check.status = 'PASS'
        check.description = 'Downward cluster resize triggered'
    return check


@check_function()
def scale_up_elasticsearch_production(connection, **kwargs):
    """ Scales up Elasticsearch (production configuration).
        HOT (0600 to 2000 EST):  This is what we are resizing to
            Master:
                3x c5.large.elasticsearch
            Data:
                2x c5.2xlarge.elasticsearch
        COLD (2000 to 0600 EST):
            Master:
                None
            Data:
                2x c5.large.elasticsearch
        XXX: should probably use constants in ElasticSearchServiceClient
        For now, must be explicitly triggered - but should be put on a schedule.
    """
    check = CheckResult(connection, 'scale_up_elasticsearch_production')
    es_client = es_utils.ElasticSearchServiceClient()
    success = es_client.resize_elasticsearch_cluster(
                domain_name=resolve_es_domain_name(connection),
                master_node_type='c5.large.elasticsearch',
                master_node_count=3,
                data_node_type='c5.2xlarge.elasticsearch',
                data_node_count=2
            )
    if not success:
        check.status = 'ERROR'
        check.description = 'Could not trigger cluster resize - check lambda logs'
    else:
        check.status = 'PASS'
        check.description = 'Downward cluster resize triggered'
    return check


# @check_function()
# def elastic_beanstalk_health(connection, **kwargs):
#     """
#     Check both environment health and health of individual instances
#     """
#     check = CheckResult(connection, 'elastic_beanstalk_health')
#     full_output = {}
#     eb_client = boto3.client('elasticbeanstalk')
#     resp = eb_client.describe_environment_health(
#         EnvironmentName=connection.ff_env,
#         AttributeNames=['All']
#     )
#     resp_status = resp.get('ResponseMetadata', {}).get('HTTPStatusCode', None)
#     if resp_status >= 400:
#         check.status = 'ERROR'
#         check.description = 'Could not establish a connection to AWS (status %s).' % resp_status
#         return check
#     full_output['status'] = resp.get('Status')
#     full_output['environment_name'] = resp.get('EnvironmentName')
#     full_output['color'] = resp.get('Color')
#     full_output['health_status'] = resp.get('HealthStatus')
#     full_output['causes'] = resp.get('Causes')
#     full_output['instance_health'] = []
#     # now look at the individual instances
#     resp = eb_client.describe_instances_health(
#         EnvironmentName=connection.ff_env,
#         AttributeNames=['All']
#     )
#     resp_status = resp.get('ResponseMetadata', {}).get('HTTPStatusCode', None)
#     if resp_status >= 400:
#         check.status = 'ERROR'
#         check.description = 'Could not establish a connection to AWS (status %s).' % resp_status
#         return check
#     instances_health = resp.get('InstanceHealthList', [])
#     for instance in instances_health:
#         inst_info = {}
#         inst_info['deploy_status'] = instance['Deployment']['Status']
#         inst_info['deploy_version'] = instance['Deployment']['VersionLabel']
#         # get version deployment time
#         application_versions = eb_client.describe_application_versions(
#             ApplicationName='4dn-web',
#             VersionLabels=[inst_info['deploy_version']]
#         )
#         deploy_info = application_versions['ApplicationVersions'][0]
#         inst_info['version_deployed_at'] = datetime.datetime.strftime(deploy_info['DateCreated'], "%Y-%m-%dT%H:%M:%S")
#         inst_info['instance_deployed_at'] = datetime.datetime.strftime(instance['Deployment']['DeploymentTime'], "%Y-%m-%dT%H:%M:%S")
#         inst_info['instance_launced_at'] = datetime.datetime.strftime(instance['LaunchedAt'], "%Y-%m-%dT%H:%M:%S")
#         inst_info['id'] = instance['InstanceId']
#         inst_info['color'] = instance['Color']
#         inst_info['health'] = instance['HealthStatus']
#         inst_info['causes'] = instance.get('causes', [])
#         full_output['instance_health'].append(inst_info)
#     if full_output['color'] == 'Grey':
#         check.status = 'WARN'
#         check.summary = check.description = 'EB environment is updating'
#     elif full_output['color'] == 'Yellow':
#         check.status = 'WARN'
#         check.summary = check.description = 'EB environment is compromised; requests may fail'
#     elif full_output['color'] == 'Red':
#         check.status = 'FAIL'
#         check.summary = check.description = 'EB environment is degraded; requests are likely to fail'
#     else:
#         check.summary = check.description = 'EB environment seems healthy'
#         check.status = 'PASS'
#     check.full_output = full_output
#     return check


@check_function()
def status_of_elasticsearch_indices(connection, **kwargs):
    check = CheckResult(connection, 'status_of_elasticsearch_indices')
    ### the check
    client = es_utils.create_es_client(connection.ff_es, True)
    indices = client.cat.indices(v=True).split('\n')
    split_indices = [ind.split() for ind in indices]
    headers = split_indices.pop(0)
    index_info = {} # for full output
    warn_index_info = {} # for brief output
    for index in split_indices:
        if len(index) == 0:
            continue
        index_info[index[2]] = {header: index[idx] for idx, header in enumerate(headers)}
        if index_info[index[2]]['health'] != 'green' or index_info[index[2]]['status'] != 'open':
            warn_index_info[index[2]] = index_info[index[2]]
    # set fields, store result
    if not index_info:
        check.status = 'FAIL'
        check.summary = 'Error reading status of ES indices'
        check.description = 'Error reading status of ES indices'
    elif warn_index_info:
        check.status = 'WARN'
        check.summary = 'ES indices may not be healthy'
        check.description = 'One or more ES indices have health != green or status != open.'
        check.brief_output = warn_index_info
    else:
        check.status = 'PASS'
        check.summary = 'ES indices seem healthy'
    check.full_output = index_info
    return check


@check_function()
def indexing_progress(connection, **kwargs):
    check = CheckResult(connection, 'indexing_progress')
    # get latest and db/es counts closest to 10 mins ago
    counts_check = CheckResult(connection, 'item_counts_by_type')
    latest = counts_check.get_primary_result()
    prior = counts_check.get_closest_result(diff_mins=30)
    if not latest.get('full_output') or not prior.get('full_output'):
        check.status = 'ERROR'
        check.description = 'There are no item_counts_by_type results to run this check with.'
        return check
    latest_unindexed = latest['full_output']['ALL']['DB'] - latest['full_output']['ALL']['ES']
    prior_unindexed = prior['full_output']['ALL']['DB'] - prior['full_output']['ALL']['ES']
    diff_unindexed = latest_unindexed - prior_unindexed
    if diff_unindexed == 0 and latest_unindexed != 0:
        check.status = 'FAIL'
        check.summary = 'Indexing is not progressing'
        check.description = ' '.join(['Total number of unindexed items is',
            str(latest_unindexed), 'and has not changed in the past thirty minutes.',
            'The indexer may be malfunctioning.'])
    elif diff_unindexed > 0:
        check.status = 'PASS'
        check.summary = 'Indexing load has increased'
        check.description = ' '.join(['Total number of unindexed items has increased by',
            str(diff_unindexed), 'in the past thirty minutes. Remaining items to index:',
            str(latest_unindexed)])
    else:
        check.status = 'PASS'
        check.summary = 'Indexing seems healthy'
        check.description = ' '.join(['Indexing seems healthy. There are', str(latest_unindexed),
        'remaining items to index, a change of', str(diff_unindexed), 'from thirty minutes ago.'])
    return check


@check_function()
def indexing_records(connection, **kwargs):
    check = CheckResult(connection, 'indexing_records')
    client = es_utils.create_es_client(connection.ff_es, True)
    namespaced_index = connection.ff_env + 'indexing'
    # make sure we have the index and items within it
    if (not client.indices.exists(namespaced_index) or
        client.count(index=namespaced_index).get('count', 0) < 1):
        check.summary = check.description = 'No indexing records found'
        check.status = 'PASS'
        return check

    res = client.search(index=namespaced_index, doc_type='indexing', sort='uuid:desc', size=1000,
                        body={'query': {'query_string': {'query': '_exists_:indexing_status'}}})
    delta_days = datetime.timedelta(days=3)
    all_records = res.get('hits', {}).get('hits', [])
    recent_records = []
    warn_records = []
    for rec in all_records:
        if rec['_id'] == 'latest_indexing':
            continue
        time_diff = (datetime.datetime.utcnow() -
            datetime.datetime.strptime(rec['_id'], "%Y-%m-%dT%H:%M:%S.%f"))
        if time_diff < delta_days:
            body = rec['_source']
            # needed to handle transition to queue. can use 'indexing_started'
            body['timestamp'] = rec['_id']
            if body.get('errors') or body.get('indexing_status') != 'finished':
                warn_records.append(body)
            recent_records.append(body)
    del all_records
    # sort so most recent records are first
    sort_records = sorted(recent_records, key=lambda rec: datetime.datetime.strptime(rec['timestamp'], "%Y-%m-%dT%H:%M:%S.%f"), reverse=True)
    check.full_output = sort_records
    if warn_records:
        sort_warn_records = sorted(warn_records, key=lambda rec: datetime.datetime.strptime(rec['timestamp'], "%Y-%m-%dT%H:%M:%S.%f"), reverse=True)
        check.summary = check.description = 'Indexing runs in the past three days may require attention'
        check.status = 'WARN'
        check.brief_output = sort_warn_records
    else:
        check.summary = check.description = 'Indexing runs from the past three days seem normal'
        check.status = 'PASS'
    return check


# this is a dummy check that is not run but instead updated with put API
# do_not_store parameter ensures running this check normally won't add to s3
@check_function(do_not_store=True)
def staging_deployment(connection, **kwargs):
    check = CheckResult(connection, 'staging_deployment')
    return check


#@check_function()
def fourfront_performance_metrics(connection, **kwargs):
    check = CheckResult(connection, 'fourfront_performance_metrics')
    full_output = {}  # contains ff_env, env_health, deploy_version, num instances, and performance
    performance = {}  # keyed by check_url
    # get information from elastic_beanstalk_health
    eb_check = CheckResult(connection, 'elastic_beanstalk_health')
    eb_info = eb_check.get_primary_result()['full_output']
    full_output['ff_env'] = connection.ff_env
    full_output['env_health'] = eb_info.get('health_status', 'Unknown')
    # get deploy version from the first instance
    full_output['deploy_version'] = eb_info.get('instance_health', [{}])[0].get('deploy_version', 'Unknown')
    full_output['num_instances'] = len(eb_info.get('instance_health', []))
    check_urls = [
        'counts',
        'joint-analysis-plans',
        'bar_plot_aggregations/type=ExperimentSetReplicate&experimentset_type=replicate/?field=experiments_in_set.experiment_type',
        'browse/?type=ExperimentSetReplicate&experimentset_type=replicate',
        'experiment-set-replicates/4DNESIE5R9HS/',
        'experiment-set-replicates/4DNESIE5R9HS/?datastore=database',
        'experiment-set-replicates/4DNESQWI9K2F/',
        'experiment-set-replicates/4DNESQWI9K2F/?datastore=database',
        'workflow-runs-awsem/ba50d240-5312-4aa7-b600-6b18d8230311/',
        'workflow-runs-awsem/ba50d240-5312-4aa7-b600-6b18d8230311/?datastore=database',
        'files-fastq/4DNFIX75FSJM/',
        'files-fastq/4DNFIX75FSJM/?datastore=database'
    ]
    for check_url in check_urls:
        performance[check_url] = {}
        try:
            # set timeout really high
            ff_resp = ff_utils.authorized_request(connection.ff_server + check_url,
                                                  auth=connection.ff_keys, timeout=1000)
        except Exception as e:
            performance[check_url]['error'] = str(e)
        if ff_resp and hasattr(ff_resp, 'headers') and 'X-stats' in ff_resp.headers:
            x_stats = ff_resp.headers['X-stats']
            if not isinstance(x_stats, str):
                performance[check_url]['error'] = 'Stats response is not a string.'
                continue
            # X-stats in form: 'db_count=148&db_time=1215810&es_count=4& ... '
            split_stats = x_stats.strip().split('&')
            parse_stats = [stat.split('=') for stat in split_stats]
            # stats can be strings or integers
            for stat in parse_stats:
                if not len(stat) == 2:
                    continue
                try:
                    performance[check_url][stat[0]] = int(stat[1])
                except ValueError:
                    performance[check_url][stat[0]] = stat[1]
            performance[check_url]['error'] = ''
    check.status = 'PASS'
    full_output['performance'] = performance
    check.full_output = full_output
    return check


#@check_function(time_limit=480)
def secondary_queue_deduplication(connection, **kwargs):
    check = CheckResult(connection, 'secondary_queue_deduplication')
    # maybe handle this in check_setup.json
    if Stage.is_stage_prod() is False:
        check.full_output = 'Will not run on dev foursight.'
        check.status = 'PASS'
        return check

    client = boto3.client('sqs')
    sqs_res = client.get_queue_url(
        QueueName=connection.ff_env + '-secondary-indexer-queue'
    )
    queue_url = sqs_res['QueueUrl']
    # get approx number of messages
    attrs = client.get_queue_attributes(
        QueueUrl=queue_url,
        AttributeNames=['ApproximateNumberOfMessages']
    )
    visible = attrs.get('Attributes', {}).get('ApproximateNumberOfMessages', '0')
    starting_count = int(visible)
    time_limit = kwargs['time_limit']
    t0 = time.time()
    sent = 0
    deleted = 0
    deduplicated = 0
    total_msgs = 0
    replaced = 0
    repeat_replaced = 0
    problem_msgs = []
    elapsed = round(time.time() - t0, 2)
    failed = []
    seen_uuids = set()
    # this is a bit of a hack -- send maximum sid with every message we replace
    # get the maximum sid at the start of deduplication and update it if we
    # encounter a higher sid
    max_sid_resp = ff_utils.authorized_request(connection.ff_server + 'max-sid',
                                               auth=connection.ff_keys).json()
    if max_sid_resp['status'] != 'success':
        check.status = 'FAIL'
        check.summary = 'Could not retrieve max_sid from the server'
        return check
    max_sid = max_sid_resp['max_sid']

    exit_reason = 'out of time'
    dedup_msg = 'FS dedup uuid: %s' % kwargs['uuid']
    while elapsed < time_limit:
        # end if we are spinning our wheels replacing the same uuids
        if (replaced + repeat_replaced) >= starting_count:
            exit_reason = 'starting uuids fully covered'
            break
        send_uuids = set()
        to_send = []
        to_delete = []
        recieved = client.receive_message(
            QueueUrl=queue_url,
            MaxNumberOfMessages=10,  # batch size for all sqs ops
            WaitTimeSeconds=1  # 1 second of long polling
        )
        batch = recieved.get("Messages", [])
        if not batch:
            exit_reason = 'no messages left'
            break
        for msg in batch:
            try:
                msg_body = json.loads(msg['Body'])
            except json.JSONDecodeError:
                problem_msgs.append(msg['Body'])
                continue
            total_msgs += 1
            msg_uuid = msg_body['uuid']
            # update max_sid with message sid if applicable
            if msg_body.get('sid') is not None and msg_body['sid'] > max_sid:
                max_sid = msg_body['sid']
            msg_body['sid'] = max_sid
            to_process = {
                'Id': msg['MessageId'],
                'ReceiptHandle': msg['ReceiptHandle']
            }
            # every item gets deleted; original uuids get re-sent
            to_delete.append(to_process)
            if msg_uuid in seen_uuids and msg_body.get('fs_detail', '') != dedup_msg:
                deduplicated += 1
            else:
                # don't increment replaced count if we've seen the item before
                if msg_uuid not in seen_uuids:
                    replaced += 1
                else:
                    repeat_replaced += 1
                time.sleep(0.0001)  # slight sleep for time-based Id
                # add foursight uuid stamp
                msg_body['fs_detail'] = dedup_msg
                # add a slight delay to recycled messages, so that they are
                # not available for consumption for 2 seconds
                send_info = {
                    'Id': str(int(time.time() * 1000000)),
                    'MessageBody': json.dumps(msg_body),
                    'DelaySeconds': 2
                }
                to_send.append(send_info)
                seen_uuids.add(msg_uuid)
                send_uuids.add(msg_uuid)
        if to_send:
            res = client.send_message_batch(
                QueueUrl=queue_url,
                Entries=to_send
            )
            # undo deduplication if errors are detected
            res_failed = res.get('Failed', [])
            failed.extend(res_failed)
            if res_failed:
                # handle conservatively on error and don't delete
                for uuid in send_uuids:
                    if uuid in seen_uuids:
                        seen_uuids.remove(uuid)
                        replaced -= 1
                continue
            sent += len(to_send)
        if to_delete:
            res = client.delete_message_batch(
                QueueUrl=queue_url,
                Entries=to_delete
            )
            failed.extend(res.get('Failed', []))
            deleted += len(to_delete)
        elapsed = round(time.time() - t0, 2)

    check.full_output = {
        'total_messages_covered': total_msgs,
        'uuids_covered': len(seen_uuids),
        'deduplicated': deduplicated,
        'replaced': replaced,
        'repeat_replaced': repeat_replaced,
        'time': elapsed,
        'problem_messages': problem_msgs,
        'exit_reason': exit_reason
    }
    # these are some standard things about the result that should always be true
    if replaced != len(seen_uuids) or (deduplicated + replaced + repeat_replaced) != total_msgs:
        check.status = 'FAIL'
        check.summary = 'Message totals do not add up. Report to Carl'
    if failed:
        if check.status != 'FAIL':
            check.status = 'WARN'
            check.summary = 'Queue deduplication encountered an error'
        check.full_output['failed'] = failed
    else:
        check.status = 'PASS'
        check.summary = 'Removed %s duplicates from %s secondary queue' % (deduplicated, connection.ff_env)
    check.description = 'Items on %s secondary queue were deduplicated. Started with approximately %s items; replaced %s items and removed %s duplicates. Covered %s unique uuids. Took %s seconds.' % (connection.ff_env, starting_count, replaced, deduplicated, len(seen_uuids), elapsed)

    return check


# @check_function()
def clean_up_travis_queues(connection, **kwargs):
    """
    Clean up old sqs queues based on the name ("travis-job")
    and the creation date. Only run on data for now
    """
    check = CheckResult(connection, 'clean_up_travis_queues')
    check.status = 'PASS'
    if connection.fs_env != 'data' or Stage.is_stage_prod() is False:
        check.summary = check.description = 'This check only runs on the data environment for Foursight prod'
        return check
    sqs_client = boto3.client('sqs')
    sqs = boto3.resource('sqs')
    queues = sqs.queues.all()
    num_deleted = 0
    for queue in queues:
        if 'travis-job' in queue.url:
            try:
                creation = queue.attributes['CreatedTimestamp']
            except sqs_client.exceptions.QueueDoesNotExist:
                continue
            if isinstance(creation, str):
                creation = float(creation)
            dt_creation = datetime.datetime.utcfromtimestamp(creation)
            queue_age = datetime.datetime.utcnow() - dt_creation
            # delete queues 3 days old or older
            if queue_age > datetime.timedelta(days=3):
                queue.delete()
                num_deleted += 1
    check.summary = 'Cleaned up %s old indexing queues' % num_deleted
    check.description = 'Cleaned up all indexing queues from Travis that are 3 days old or older. %s queues deleted.' % num_deleted
    return check


# @check_function()
# def manage_old_filebeat_logs(connection, **kwargs):
#     # import curator
#     check = CheckResult(connection, 'manage_old_filebeat_logs')
#
#     # temporary -- disable this check
#     check.status = 'PASS'
#     check.description = 'Not currently running this check'
#     return check

    # check.status = "WARNING"
    # check.description = "not able to get data from ES"
    #
    # # configure this thing
    # start_backup = 14
    # trim_backup = 30
    # timestring = '%Y.%m.%d'
    #
    # log_index = 'filebeat-'
    # #TODO: this probably needs to change when namespacing is implemented
    # snapshot = 'backup-%s-' % connection.ff_env
    # today = datetime.datetime.today().strftime(timestring)
    #
    # # run the check
    # client = es_utils.create_es_client(connection.ff_es, True)
    # # store backups in foursight s3, cause I know we have access to this..
    # # maybe this should change?
    # es_utils.create_snapshot_repo(client, snapshot[:-1], 'foursight-runs')
    #
    # # amazon es auto backups first 14 days, so we only need backup after that
    # ilo = es_utils.get_index_list(client, log_index, start_backup, timestring)
    #
    # if len(ilo.indices) > 0:
    #     try:
    #         new_snapshot = curator.Snapshot(ilo, repository=snapshot[:-1], name='%s%s' % (snapshot, today))
    #         new_snapshot.do_action()
    #         check.full_output = "Snapshot taken for %s indices" % len(ilo.indices)
    #     except curator.exceptions.FailedExecution as e:
    #         # snapshot already exists
    #         if "Invalid snapshot name" in str(e):
    #             check.full_output = "Snapshot already exists with same name for %s indices, so skipping." % len(ilo.indices)
    #         else:
    #             raise(e)
    #
    # # now trim further to only be indexes 30-days or older and delete
    # ilo = es_utils.get_index_list(client, log_index, trim_backup, timestring, ilo=ilo)
    #
    # if len(ilo.indices) > 0:
    #     cleanupIndices = curator.DeleteIndices(ilo)
    #     cleanupIndices.do_action()
    #     check.full_output += " Cleaned up %s old indices" % len(ilo.indices)
    #
    # check.status = "PASS"
    # check.description = 'Performed auto-backup to repository %s' % snapshot[:-1]
    # return check


@check_function()
def snapshot_rds(connection, **kwargs):
    check = CheckResult(connection, 'snapshot_rds')
    if Stage.is_stage_prod() is False:
        check.summary = check.description = 'This check only runs on Foursight prod'
        return check
    rds_name = 'fourfront-production' if (env_utils.is_fourfront_env(connection.ff_env) and env_utils.is_stg_or_prd_env(connection.ff_env)) else connection.ff_env
    # snapshot ID can only have letters, numbers, and hyphens
    snap_time = datetime.datetime.strptime(kwargs['uuid'], "%Y-%m-%dT%H:%M:%S.%f").strftime("%Y-%m-%dT%H-%M-%S")
    snapshot_name = 'foursight-snapshot-%s-%s' % (rds_name, snap_time)
    client = boto3.client('rds', region_name=beanstalk_utils.REGION)
    res = client.create_db_snapshot(
             DBSnapshotIdentifier=snapshot_name,
             DBInstanceIdentifier=rds_name
    )
    if not res.get('DBSnapshot'):
        check.status = 'FAIL'
        check.summary = check.description = 'Something went wrong during snapshot creation'
        check.full_output = res
    else:
        check.status = 'PASS'
        # there is a datetime in the response that must be str formatted
        res['DBSnapshot']['InstanceCreateTime'] = str(res['DBSnapshot']['InstanceCreateTime'])
        check.full_output = res
        check.summary = 'Snapshot successfully created'
        check.description = 'Snapshot succesfully created with name: %s' % snapshot_name
    return check


# @check_function()
def process_download_tracking_items(connection, **kwargs):
    """
    Do a few things here, and be mindful of the 5min lambda limit.
    - Consolidate tracking items with download_tracking.range_query=True
    - Change remote_ip to geo_country and geo_city
    - If the user_agent looks to be a bot, set status=deleted
    - Change unused range query items to status=deleted
    """
    check = CheckResult(connection, 'process_download_tracking_items')
    # maybe handle this in check_setup.json
    if Stage.is_stage_prod() is False:
        check.full_output = 'Will not run on dev foursight.'
        check.status = 'PASS'
        return check
    # hold warning messages
    check.brief_output = {'cannot_parse_date_created': []}
    range_cache = {}
    # duration we want to consolidate range queries over
    # search older entries since range_consolidation_hrs * 2 to avoid duplication
    range_consolidation_hrs = 1
    cons_date = (datetime.datetime.utcnow() -
                 datetime.timedelta(hours=range_consolidation_hrs * 2)).strftime('%Y-%m-%dT%H\:%M')
    range_search_query = ''.join(['search/?type=TrackingItem&tracking_type=download_tracking',
                                  '&download_tracking.range_query=true&sort=-date_created',
                                  '&status=released&q=last_modified.date_modified:>=', cons_date])
    cons_query = ff_utils.search_metadata(range_search_query, key=connection.ff_keys)
    for tracking in cons_query:
        dl_info = tracking['download_tracking']
        user_agent = dl_info.get('user_agent', 'unknown_user_agent').lower()
        range_key = '//'.join([dl_info['remote_ip'], dl_info['filename'],
                               user_agent, dl_info['user_uuid']])
        parsed_date = parse_datetime_to_utc(tracking['date_created'])
        # check for date parsing error
        if parsed_date is None:
            continue
        if range_key in range_cache:
            range_cache[range_key].append(parsed_date)
        else:
            range_cache[range_key] = [parsed_date]
    del cons_query
    ip_cache = {}
    time_limit = 270  # 4.5 minutes
    # list of strings used to flag user_agent as a bot. By no means complete
    bot_agents = ['bot', 'crawl', 'slurp', 'spider', 'mediapartners', 'ltx71']
    # list of user_agents we always consider range_query=True
    range_query_agents = ['igv', 'java', 'python-requests']
    t0 = time.time()  # keep track of how start time

    # batch large groups of tracking items at once to save time with geocoder
    # for now, this function will process only <search_limit> results.
    # I would love to use a generator, but search results change as items are indexed...
    search_limit = 1000
    search_query = ''.join(['search/?type=TrackingItem&tracking_type=download_tracking',
                            '&download_tracking.geo_country=No+value',
                            '&status=in+review+by+lab&sort=-date_created&limit=', str(search_limit)])
    search_page = ff_utils.search_metadata(search_query, key=connection.ff_keys, page_limit=200)
    counts = {'proc': 0, 'deleted': 0, 'released': 0}

    page_ips = set([tracking['download_tracking']['remote_ip'] for tracking in search_page])
    # transform all IP addresses into GEO information with a persistent connection
    with requests.Session() as session:
        for track_ip in page_ips:
            if track_ip in ip_cache:
                continue
            geo = geocoder.ip(track_ip, session=session)
            geo_country = getattr(geo, 'country') or 'Unknown'
            geo_city = getattr(geo, 'city') or 'Unknown'
            geo_state = getattr(geo, 'state', None)
            if geo_state:
                geo_city = ', '.join([geo_city, geo_state])
            # cache the geo info in an arbitrary form
            ip_cache[track_ip] = '//'.join([geo_city, geo_country])

    # iterate over the individual tracking items
    for tracking in search_page:
        if round(time.time() - t0, 2) > time_limit:
            break
        dl_info = tracking['download_tracking']
        user_agent = dl_info.get('user_agent', 'unknown_user_agent').lower()
        # remove request_headers, which may contain sensitive information
        if 'request_headers' in dl_info:
            del dl_info['request_headers']
        geo_info = ip_cache[dl_info['remote_ip']]
        dl_info['geo_city'], dl_info['geo_country'] = geo_info.split('//')
        patch_body = {'status': 'released', 'download_tracking': dl_info}
        # delete items from bot user agents
        if (any(bot_str in user_agent for bot_str in bot_agents)
            and dl_info['user_uuid'] == 'anonymous'):
            patch_body['status'] = 'deleted'
        # set range_query=True for select user agents
        if (any(ua_str in user_agent for ua_str in range_query_agents)):
            dl_info['range_query'] = True
        # deduplicate range query requests by ip/filename/user_agent/user_uuid
        if patch_body['status'] != 'deleted' and dl_info['range_query'] is True:
            range_key = '//'.join([dl_info['remote_ip'], dl_info['filename'],
                                   user_agent, dl_info['user_uuid']])
            parsed_date = parse_datetime_to_utc(tracking['date_created'])
            if parsed_date is not None:
                if range_key in range_cache:
                    # for all reference range queries with this info, see if this one
                    # was created within one hour of it. if so, it is redundant and delete
                    for range_reference in range_cache[range_key]:
                        compare_date_low = range_reference - datetime.timedelta(hours=range_consolidation_hrs)
                        compare_date_high = range_reference + datetime.timedelta(hours=range_consolidation_hrs)
                        if parsed_date > compare_date_low and parsed_date < compare_date_high:
                            patch_body['status'] = 'deleted'
                            break
                    if patch_body['status'] != 'deleted':
                        range_cache[range_key].append(parsed_date)
                else:
                    # set the upper limit for for range queries to consolidate
                    range_cache[range_key] = [parsed_date]
            else:
                check.brief_output['cannot_parse_date_created'].append(tracking['uuid'])

        ff_utils.patch_metadata(patch_body, tracking['uuid'], key=connection.ff_keys)
        counts['proc'] += 1
        if patch_body['status'] == 'released':
            counts['released'] += 1
        else:
            counts['deleted'] += 1
    if any(check.brief_output.values()):
        check.status = 'WARN'
    else:
        check.status = 'PASS'
    check.summary = 'Successfully processed %s download tracking items' % counts['proc']
    check.description = '%s. Released %s items and deleted %s items' % (check.summary, counts['released'], counts['deleted'])
    return check


# @check_function()
def purge_download_tracking_items(connection, **kwargs):
    """
    This check was originally created to take in any search through kwargs.
    Changed to hardcode a search for tracking items, but it can easily
    adapted; as it is, already handles recording for any number of item types.
    Ensure search includes limit, field=uuid, and status=deleted
    """
    check = CheckResult(connection, 'purge_download_tracking_items')

    # Don't run if staging deployment is running
    # Only need to check if our env is data
    # XXX: Removing for now as we find the check can never run without this
    # if the staging deploy takes long enough or errors
    # if connection.fs_env == 'data':
    #     from ..app_utils import AppUtils
    #     staging_conn = AppUtils().init_connection('staging')
    #     staging_deploy = CheckResult(staging_conn, 'staging_deployment').get_primary_result()
    #     if staging_deploy['status'] != 'PASS':
    #         check.summary = 'Staging deployment is running - skipping'
    #         return check

    if Stage.is_stage_prod() is False:
        check.summary = check.description = 'This check only runs on Foursight prod'
        return check

    time_limit = 270  # 4.5 minutes
    t0 = time.time()
    check.full_output = {}  # purged items by item type
    search = '/search/?type=TrackingItem&tracking_type=download_tracking&status=deleted&field=uuid&limit=300'
    search_res = ff_utils.search_metadata(search, key=connection.ff_keys)
    search_uuids = [res['uuid'] for res in search_res]
    client = es_utils.create_es_client(connection.ff_es, True)
    # a bit convoluted, but we want the frame=raw, which does not include uuid
    # use get_es_metadata to handle this. Use it as a generator
    for to_purge in ff_utils.get_es_metadata(search_uuids, es_client=client, is_generator=True,
                                             key=connection.ff_keys):
        if round(time.time() - t0, 2) > time_limit:
            break
        purge_properties = to_purge['properties']
        purge_properties['uuid'] = to_purge['uuid']  # add uuid to frame=raw
        try:
            purge_res = ff_utils.purge_metadata(to_purge['uuid'], key=connection.ff_keys)
        except Exception as exc:
            purge_status = 'error'
            purge_detail = str(exc)
        else:
            purge_status = purge_res['status']
            purge_detail = purge_properties if purge_status == 'success' else purge_res
        purge_record = {'uuid': to_purge['uuid'], 'result': purge_detail}
        if to_purge['item_type'] not in check.full_output:
            check.full_output[to_purge['item_type']] = {}
        if purge_status not in check.full_output[to_purge['item_type']]:
            check.full_output[to_purge['item_type']][purge_status] = []
        check.full_output[to_purge['item_type']][purge_status].append(purge_record)
    purge_out_str = '. '.join(['%s: %s' % (it, len(check.full_output[it]['success']))
                               for it in check.full_output if check.full_output[it].get('success')])
    check.description = 'Purged: ' + purge_out_str + '. Search used: %s' % search
    if any([it for it in check.full_output if check.full_output[it].get('error')]):
        check.status = 'WARN'
        check.summary = 'Some items failed to purge. See full output'
    else:
        check.status = 'PASS'
        check.summary = 'Items purged successfully'
    return check


@check_function()
def check_long_running_ec2s(connection, **kwargs):
    """
    Flag all ec2s that have been running for longer than 1 week (WARN) or 2 weeks
    (FAIL) if any contain any strings from `flag_names` in their
    names, or if they have no name.
    """
    check = CheckResult(connection, 'check_long_running_ec2s')
    if Stage.is_stage_prod() is False:
        check.summary = check.description = 'This check only runs on Foursight prod'
        return check

    client = boto3.client('ec2')
    # flag instances that contain any of flag_names and have been running
    # longer than warn_time
    flag_names = ['awsem']
    warn_time = (datetime.datetime.now(datetime.timezone.utc) -
                 datetime.timedelta(days=7))
    fail_time = (datetime.datetime.now(datetime.timezone.utc) -
                 datetime.timedelta(days=14))
    ec2_res = client.describe_instances(
        Filters=[{'Name': 'instance-state-name', 'Values': ['running']}]
    )
    check.full_output = []
    check.brief_output = {'one_week': [], 'two_weeks': []}
    for ec2_info in ec2_res.get('Reservations', []):
        instances = ec2_info.get('Instances', [])
        if not instances:
            continue
        # for multiple instance (?) just check if any of them require warnings
        for ec2_inst in instances:
            state = ec2_inst.get('State')
            created = ec2_inst.get('LaunchTime')
            if not state or not created:
                continue
            inst_name = [kv['Value'] for kv in ec2_inst.get('Tags', [])
                         if kv['Key'] == 'Name']
            other_tags = {kv['Key']: kv['Value'] for kv in ec2_inst.get('Tags', [])
                         if kv['Key'] != 'Name'}
            ec2_log = {
                'state': state['Name'], 'name': inst_name,
                'id': ec2_inst.get('InstanceId'),
                'type': ec2_inst.get('InstanceType'),
                'date_created_utc': created.strftime('%Y-%m-%dT%H:%M')
            }
            if not inst_name:
                flag_instance = True
                # include all other tags if Name tag is empty
                ec2_log['tags'] = other_tags
            elif any([wn for wn in flag_names if wn in ','.join(inst_name)]):
                flag_instance = True
            else:
                flag_instance = False
            # see if long running instances are associated with a deleted WFR
            if flag_instance and inst_name and created < warn_time:
                search_url = 'search/?type=WorkflowRunAwsem&awsem_job_id='
                search_url += '&awsem_job_id='.join([name[6:] for name in inst_name if name.startswith('awsem-')])
                wfrs = ff_utils.search_metadata(search_url, key=connection.ff_keys)
                if wfrs:
                    ec2_log['active workflow runs'] = [wfr['@id'] for wfr in wfrs]
                deleted_wfrs = ff_utils.search_metadata(search_url + '&status=deleted', key=connection.ff_keys)
                if deleted_wfrs:
                    ec2_log['deleted workflow runs'] = [wfr['@id'] for wfr in deleted_wfrs]
            # always add record to full_output; add to brief_output if
            # the instance is flagged based on 'Name' tag
            if created < fail_time:
                if flag_instance:
                    check.brief_output['two_weeks'].append(ec2_log)
                check.full_output.append(ec2_log)
            elif created < warn_time:
                if flag_instance:
                    check.brief_output['one_week'].append(ec2_log)
                check.full_output.append(ec2_log)

    if check.brief_output['one_week'] or check.brief_output['two_weeks']:
        num_1wk = len(check.brief_output['one_week'])
        num_2wk = len(check.brief_output['two_weeks'])
        check.summary = ''
        if check.brief_output['two_weeks']:
            check.status = 'FAIL'
            check.summary = '%s suspect EC2s running longer than 2 weeks' % num_2wk
        if check.brief_output['one_week']:
            if check.status != 'FAIL':
                check.status = 'WARN'
            if check.summary:
                check.summary += ' and %s others longer than 1 week' % num_1wk
            else:
                check.summary = '%s suspect EC2s running longer than 1 week' % num_1wk
        check.description = check.summary + '. Flagged because name is empty or contains %s. There are also %s non-flagged instances.' % (flag_names, len(check.full_output) - (num_1wk + num_2wk))
    else:
        check.status = 'PASS'
        check.summary = '%s EC2s running longer than 1 week' % (len(check.full_output))
    return check


@check_function(FS_dev='free', FF_hotseat='free', FF_mastertest='free', FF_webdev='free')
def say_my_name(connection, **kwargs):
    """List the person working on each environment."""
    check = CheckResult(connection, 'say_my_name')
    check.description = "Enter the new name or if you are done, use 'free' to clear your name"
    check.summary = ""
    check.brief_output = ""
    check.status = "PASS"
    output = {}
    # update with the new parameters ()
    for a_key in ['FS_dev', 'FF_hotseat', 'FF_mastertest', 'FF_webdev']:
        if kwargs.get(a_key):
            val = kwargs[a_key]
            if val.lower() == 'free':
                val = 'free'
            output[a_key] = val
        else:
            output[a_key] = 'free'
    sum = str(output)[1:-1].replace("'", "")
    check.summary = sum
    check.brief_output = sum
    return check
